import math
import torch
import torch.nn as nn
import torch.nn.functional as F

from typing import Optional


def binary_enc_matrix(num_embeddings: int):
    bits = math.ceil(math.log2(num_embeddings))
    mask = 2 ** torch.arange(bits)
    x = torch.arange(mask.sum().item() + 1)
    embs = x.unsqueeze(-1).bitwise_and(mask).ne(0).float()
    return embs[..., :num_embeddings:, :]


def sqrt_one_hot_matrix(num_embeddings: int):
    x = torch.arange(num_embeddings)
    i = (x**0.5).to(torch.long)
    return F.one_hot(i, torch.max(i) + 1).to(torch.float)


def power_one_hot_matrix(num_embeddings: int, power: float):
    x = torch.arange(num_embeddings)
    i = (x**power).to(torch.long)
    return F.one_hot(i, torch.max(i) + 1).to(torch.float)


def _legal_policy(logits: torch.Tensor, legal_actions: torch.Tensor) -> torch.Tensor:
    """A soft-max policy that respects legal_actions."""
    # Fiddle a bit to make sure we don't generate NaNs or Inf in the middle.
    policy = torch.masked_fill(logits, ~legal_actions, float("-inf")).softmax(-1)
    return policy


def _log_policy(logits: torch.Tensor, legal_actions: torch.Tensor) -> torch.Tensor:
    return F.log_softmax(torch.masked_fill(logits, ~legal_actions, -1e9), dim=-1)


def gather_along_rows(
    x: torch.Tensor, index: torch.Tensor, dim: int = 1, keepdim: bool = True
):
    """
    x: torch.Tensor -> [B, T, D]
    dim: int
    index: torch.Tensor -> [B,], values in range (0, T)
    """
    assert len(x.shape) == 3
    B, _, D = x.shape
    index = index.view(B, 1).repeat(1, D).view(B, 1, D)
    rows = torch.gather(x, dim, index)
    if keepdim:
        return rows
    else:
        return rows.squeeze(dim)


class GLU(nn.Module):
    def __init__(self, input_dim: int, output_dim: int, context_dim: int):
        super().__init__()
        self.lin1 = nn.Linear(context_dim, input_dim)
        self.lin2 = nn.Linear(input_dim, output_dim)

    def forward(self, x: torch.Tensor, context: torch.Tensor):
        gate = self.lin1(context)
        gate = torch.sigmoid(gate)
        x = gate * x
        x = self.lin2(x)
        return x


class Resblock(nn.Module):
    def __init__(
        self,
        hidden_size: int,
        num_layers: int = 2,
        use_layer_norm: bool = True,
    ) -> None:
        super().__init__()

        layers = []
        for _ in range(num_layers):
            hiddens = [nn.ReLU(), nn.Linear(hidden_size, hidden_size)]
            if use_layer_norm:
                hiddens = [nn.LayerNorm(hidden_size)] + hiddens
            layers.append(nn.Sequential(*hiddens))
        self.layers = nn.ModuleList(layers)

    def forward(self, x: torch.Tensor):
        shortcut = x
        for layer in self.layers:
            x = layer(x)
        return x + shortcut


class MultiHeadAttention(nn.Module):
    def __init__(
        self,
        input_size: int,
        num_heads: int,
        key_size: int,
        value_size: int = None,
        model_size: int = None,
    ):
        super().__init__()

        self.num_heads = num_heads
        self.key_size = key_size
        value_size = value_size or key_size
        model_size = model_size or key_size * num_heads
        self.value_size = value_size
        self.model_size = model_size

        self.lin_q = nn.Linear(input_size, num_heads * key_size)
        self.lin_k = nn.Linear(input_size, num_heads * key_size)
        self.lin_v = nn.Linear(input_size, num_heads * value_size)
        self.lin_out = nn.Linear(num_heads * value_size, model_size)

        self.attn_denom = key_size**0.5

    def _linear_projection(
        self, x: torch.Tensor, layer: nn.Module, head_size: int
    ) -> torch.Tensor:
        y = layer(x)
        *leading_dims, _ = x.shape
        return y.reshape((*leading_dims, self.num_heads, head_size))

    def forward(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        mask: torch.Tensor,
    ):
        *leading_dims, sequence_length, _ = query.shape

        query_heads = self._linear_projection(query, self.lin_q, self.key_size)
        key_heads = self._linear_projection(key, self.lin_k, self.key_size)
        value_heads = self._linear_projection(value, self.lin_v, self.value_size)

        attn_logits = torch.einsum("...thd,...Thd->...htT", query_heads, key_heads)
        attn_logits = attn_logits / self.attn_denom

        if mask is not None:
            if mask.ndim != attn_logits.ndim:
                raise ValueError(
                    f"Mask dimensionality {mask.ndim} must match logits dimensionality "
                    f"{attn_logits.ndim}."
                )
            attn_logits = torch.where(mask, attn_logits, float("-inf"))
        attn_weights = torch.softmax(attn_logits, dim=-1)  # [H, T', T]

        # Weight the values by the attention and flatten the head vectors.
        attn = torch.einsum("...htT,...Thd->...thd", attn_weights, value_heads)
        attn = torch.reshape(attn, (*leading_dims, sequence_length, -1))  # [T', H*V]

        # Apply another projection to get the final embeddings.
        return self.lin_out(attn)  # [T', D']


class TransformerLayer(nn.Module):
    def __init__(
        self,
        transformer_num_heads: int,
        transformer_key_size: int,
        transformer_value_size: int,
        input_size: int,
        model_size: int,
        use_layer_norm: bool = True,
    ) -> None:
        super().__init__()
        self._use_layer_norm = use_layer_norm
        if use_layer_norm:
            self.ln = nn.LayerNorm(model_size)
        self.mha = MultiHeadAttention(
            num_heads=transformer_num_heads,
            key_size=transformer_key_size,
            value_size=transformer_value_size,
            model_size=model_size,
            input_size=input_size,
        )

    def forward(self, x: torch.Tensor, mask: torch.Tensor):
        x1 = x
        if self._use_layer_norm:
            x1 = self.ln(x1)

        x1 = F.relu(x1)
        # The logits mask has shape [num_heads, num_units, num_units]:
        x1 = self.mha(x1, x1, x1, mask[:, None, None, :])

        # Mask here mostly for safety:
        x1 = torch.where(mask.unsqueeze(-1), x1, 0)
        x = x + x1

        return x


class TransformerEncoder(nn.Module):
    def __init__(
        self,
        model_size: int,
        num_layers: int,
        num_heads: int,
        key_size: int,
        value_size: int,
        resblocks_num_before: int,
        resblocks_num_after: int,
        resblocks_hidden_size: Optional[int] = None,
        use_layer_norm: bool = True,
    ):

        super().__init__()
        self._transformer_num_layers = num_layers
        self._transformer_num_heads = num_heads
        self._transformer_key_size = key_size
        self._transformer_value_size = value_size

        if resblocks_hidden_size is None:
            resblocks_hidden_size = model_size

        self._transformer_layers = nn.ModuleList(
            [
                TransformerLayer(
                    num_heads,
                    key_size,
                    value_size,
                    input_size=resblocks_hidden_size,
                    model_size=model_size,
                    use_layer_norm=use_layer_norm,
                )
                for _ in range(num_layers)
            ]
        )

        self._resblocks_before = nn.ModuleList(
            [
                Resblock(
                    hidden_size=resblocks_hidden_size,
                    use_layer_norm=use_layer_norm,
                )
                for _ in range(resblocks_num_before)
            ]
        )
        self._resblocks_after = nn.ModuleList(
            [
                Resblock(
                    hidden_size=resblocks_hidden_size,
                    use_layer_norm=use_layer_norm,
                )
                for _ in range(resblocks_num_after)
            ]
        )

        self._resblocks_hidden_size = resblocks_hidden_size
        self._use_layer_norm = use_layer_norm

    def forward(self, x: torch.Tensor, mask: torch.Tensor = None) -> torch.Tensor:
        for resblock_before in self._resblocks_before:
            x = resblock_before(x)

        for transformer_layer in self._transformer_layers:
            x = transformer_layer(x, mask)

        for resblock_after in self._resblocks_after:
            x = resblock_after(x)

        # Mask here mostly for safety:
        x = torch.where(mask.unsqueeze(-1), x, 0)
        return x


class AttentionPool(nn.Module):
    def __init__(self, embed_dim: int, nhead: int) -> None:
        super().__init__()

        self.nhead = nhead
        self.lin_qk = nn.Linear(embed_dim, 3 * embed_dim)
        self.proj = nn.Linear(embed_dim, embed_dim)

    def forward(self, x: torch.Tensor, mask: torch.Tensor):
        TB, S, C = x.shape

        q, k, v = torch.chunk(self.lin_qk(x), 3, 2)
        q = q.view(TB, S, self.nhead, C // self.nhead).transpose(1, 2)
        k = k.view(TB, S, self.nhead, C // self.nhead).transpose(1, 2)
        v = v.view(TB, S, self.nhead, C // self.nhead).transpose(1, 2)

        attn = (q @ k.transpose(-2, -1)) * (1.0 / math.sqrt(k.size(-1)))
        mask = (mask * mask.transpose(-2, -1)).unsqueeze(1)
        attn = torch.masked_fill(attn, mask, float("-inf"))
        score = F.softmax(attn, -1)

        y = score @ v
        y = y.transpose(1, 2).contiguous().view(TB, S, C)
        y = y.sum(1, keepdim=True)
        y = self.proj(y)

        return y


class ToVector(nn.Module):
    """Per-unit processing then average over the units dimension."""

    def __init__(self, input_dim: int, output_dim: int):
        super().__init__()

        self.ln = nn.LayerNorm(2 * input_dim)
        self.lin = nn.Linear(2 * input_dim, output_dim)

    def forward(self, x: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        x = torch.cat(
            (
                x[..., :1, :].sum(1) / mask[..., :1, None].sum(1).clamp(min=1),
                x[..., 1:, :].sum(1) / mask[..., 1:, None].sum(1).clamp(min=1),
            ),
            dim=-1,
        )
        x = self.ln(x)
        x = F.relu(x)
        x = self.lin(x)
        return x



import torch
import torch.nn as nn

from meloetta.frameworks.nash_ketchum.model import config
from meloetta.frameworks.nash_ketchum.model.utils import (
    _legal_policy,
    _multinomial,
    MLP,
    BIAS,
)

from meloetta.data import CHOICE_FLAGS


class FlagsHead(nn.Module):
    def __init__(self, config: config.FlagHeadConfig):
        super().__init__()

        self.mlp = MLP(
            [
                config.autoregressive_embedding_dim,
                config.autoregressive_embedding_dim,
                len(CHOICE_FLAGS),
            ]
        )
        self.proj_flag = MLP(
            [
                len(CHOICE_FLAGS),
                config.autoregressive_embedding_dim,
                config.autoregressive_embedding_dim,
            ]
        )
        self.one_hot = nn.Embedding.from_pretrained(torch.eye(len(CHOICE_FLAGS)))

    def forward(
        self,
        action_index: torch.Tensor,
        autoregressive_embedding: torch.Tensor,
        flag_mask: torch.Tensor,
        flag_index: torch.Tensor = None,
    ):
        T, B, *_ = autoregressive_embedding.shape

        flag_logits = self.mlp(autoregressive_embedding)
        flag_policy = _legal_policy(flag_logits, flag_mask)
        flag_logits = torch.where(flag_mask, flag_logits, BIAS)
        flag_policy = flag_policy.view(T * B, -1)

        if flag_index is None:
            flag_index = _multinomial(flag_policy)

        flag_policy = flag_policy.view(T, B, -1)
        flag_index = flag_index.view(T, B, -1)

        flag_one_hot = self.one_hot(flag_index.long())
        projected_flag_embedding = self.proj_flag(flag_one_hot)
        projected_flag_embedding = projected_flag_embedding.view(T, B, -1)

        valid_indices = action_index == 0
        valid_indices = valid_indices.view(T, B, 1)

        autoregressive_embedding = (
            autoregressive_embedding + valid_indices * projected_flag_embedding
        )

        return (
            flag_logits,
            flag_policy,
            flag_index,
            autoregressive_embedding,
        )



import torch
import torch.nn as nn
import torch.nn.functional as F

from meloetta.frameworks.porygon.model import config
from meloetta.frameworks.porygon.model.utils import Resblock
from meloetta.frameworks.porygon.model.utils import _legal_policy


class ActionTypeHead(nn.Module):
    def __init__(self, config: config.ActionTypeHeadConfig):
        super().__init__()

        self.num_action_types = config.num_action_types
        self.project = nn.Linear(config.state_embedding_dim, config.residual_dim)
        self.resblocks = nn.Sequential(
            *[Resblock(config.residual_dim, use_layer_norm=True) for _ in range(1)]
        )

        self.action_fc = nn.Linear(config.residual_dim, config.num_action_types)

        self.action_type_emb = nn.Embedding(3, config.state_embedding_dim)

    def forward(self, state: torch.Tensor, action_type_mask: torch.Tensor):
        T, B, *_ = state.shape

        embedded_state = self.project(state)
        embedded_state = F.relu(self.resblocks(embedded_state))
        action_type_logits = self.action_fc(embedded_state)
        action_type_policy = _legal_policy(action_type_logits, action_type_mask)
        action_type_index = torch.multinomial(action_type_policy.view(T * B, -1), 1)
        action_type_index = action_type_index.view(T, B)

        action_type_embedding = self.action_type_emb(action_type_index.long())

        autoregressive_embedding = state + action_type_embedding

        return (
            action_type_logits,
            action_type_policy,
            action_type_index,
            autoregressive_embedding,
        )



import torch
import torch.nn as nn

from meloetta.frameworks.nash_ketchum.model.utils import Resblock

from meloetta.frameworks.nash_ketchum.model import config


class ValueHead(nn.Module):
    def __init__(self, config: config.ValueHeadConfig):
        super().__init__()
        self.lin_in = nn.Linear(config.state_embedding_dim, config.hidden_dim)

        self.resblock_stack = nn.ModuleList(
            [
                Resblock(config.hidden_dim, use_layer_norm=False)
                for _ in range(config.num_resblocks)
            ]
        )
        self.lin_out = nn.Linear(config.hidden_dim, 1)

    def forward(self, x: torch.Tensor):
        x = self.lin_in(x)
        for resblock in self.resblock_stack:
            x = resblock(x)
        return self.lin_out(x)



import math
import torch
import collections

from typing import NamedTuple, Iterator, Dict, List

from meloetta.actors.types import State, Choices


class SideEncoderOutput(NamedTuple):
    side_embedding: torch.Tensor
    side_context: torch.Tensor
    targ_hidden: torch.Tensor
    pred_hidden: torch.Tensor
    moves: torch.Tensor
    switches: torch.Tensor


class EncoderOutput(NamedTuple):
    side_embs: SideEncoderOutput
    weather_emb: torch.Tensor
    scalar_emb: torch.Tensor


class Indices(NamedTuple):
    action_type_index: torch.Tensor
    move_index: torch.Tensor
    max_move_index: torch.Tensor
    switch_index: torch.Tensor
    flag_index: torch.Tensor
    target_index: torch.Tensor


class Logits(NamedTuple):
    action_type_logits: torch.Tensor = None
    move_logits: torch.Tensor = None
    max_move_logits: torch.Tensor = None
    switch_logits: torch.Tensor = None
    flag_logits: torch.Tensor = None
    target_logits: torch.Tensor = None


class Policy(NamedTuple):
    action_type_policy: torch.Tensor = None
    move_policy: torch.Tensor = None
    max_move_policy: torch.Tensor = None
    switch_policy: torch.Tensor = None
    flag_policy: torch.Tensor = None
    target_policy: torch.Tensor = None


class LogPolicy(NamedTuple):
    action_type_log_policy: torch.Tensor = None
    move_log_policy: torch.Tensor = None
    max_move_log_policy: torch.Tensor = None
    switch_log_policy: torch.Tensor = None
    flag_log_policy: torch.Tensor = None
    target_log_policy: torch.Tensor = None


class TrainingOutput(NamedTuple):
    pi: Policy
    v: torch.Tensor
    logit: Logits
    log_pi: LogPolicy = None


class ModelOutput(NamedTuple):
    indices: Indices
    policy: Policy
    logits: Logits
    value: torch.Tensor

    def to_store(self, state: State) -> Dict[str, torch.Tensor]:
        to_store = {k: v for k, v in state.items() if isinstance(v, torch.Tensor)}
        to_store.update(
            {
                k: v.squeeze()
                for k, v in self.indices._asdict().items()
                if isinstance(v, torch.Tensor)
            }
        )
        to_store.update(
            {
                k: v.squeeze()
                for k, v in self.policy._asdict().items()
                if isinstance(v, torch.Tensor)
            }
        )
        return to_store


class PostProcess(NamedTuple):
    data: Choices
    index: torch.Tensor


class Batch(NamedTuple):
    sides: torch.Tensor = None
    boosts: torch.Tensor = None
    volatiles: torch.Tensor = None
    side_conditions: torch.Tensor = None
    pseudoweathers: torch.Tensor = None
    weather: torch.Tensor = None
    wisher: torch.Tensor = None
    turn: torch.Tensor = None
    n: torch.Tensor = None
    total_pokemon: torch.Tensor = None
    faint_counter: torch.Tensor = None
    turn: torch.Tensor = None
    turns_since_last_move: torch.Tensor = None
    action_type_mask: torch.Tensor = None
    move_mask: torch.Tensor = None
    max_move_mask: torch.Tensor = None
    switch_mask: torch.Tensor = None
    flag_mask: torch.Tensor = None
    rewards: torch.Tensor = None
    player_id: torch.Tensor = None
    action_type_policy: torch.Tensor = None
    move_policy: torch.Tensor = None
    max_move_policy: torch.Tensor = None
    switch_policy: torch.Tensor = None
    flag_policy: torch.Tensor = None
    action_type_index: torch.Tensor = None
    move_index: torch.Tensor = None
    max_move_index: torch.Tensor = None
    switch_index: torch.Tensor = None
    flag_index: torch.Tensor = None
    value: torch.Tensor = None
    valid: torch.Tensor = None

    def get_index_from_policy(self, policy_field: str):
        return getattr(self, policy_field.replace("policy", "index"))

    def get_mask_from_policy(self, policy_field: str):
        return getattr(self, policy_field.replace("policy", "mask"))

    @property
    def batch_size(self):
        return self.valid.shape[1]

    @property
    def trajectory_lengths(self):
        return self.valid.sum(0)

    @property
    def trajectory_length(self):
        return self.valid.shape[0]

    def slice(self, start: int, end: int, max_length: int = None) -> "Batch":
        return Batch(
            **{
                key: value[:max_length, start:end].contiguous()
                for key, value in self._asdict().items()
                if value is not None
            }
        )


class Loss(NamedTuple):
    action_type_policy_loss: float = 0
    move_policy_loss: float = 0
    switch_policy_loss: float = 0
    max_move_policy_loss: float = 0
    flag_policy_loss: float = 0
    target_policy_loss: float = 0

    action_type_value_loss: float = 0
    move_value_loss: float = 0
    switch_value_loss: float = 0
    max_move_value_loss: float = 0
    flag_value_loss: float = 0
    target_value_loss: float = 0

    action_type_entropy_loss: float = 0
    move_entropy_loss: float = 0
    switch_entropy_loss: float = 0
    max_move_entropy_loss: float = 0
    flag_entropy_loss: float = 0
    target_entropy_loss: float = 0

    recon_loss: float = 0

    def to_log(self, batch: Batch):
        logs = {
            key: value for key, value in self._asdict().items() if value is not None
        }
        logs["traj_len"] = batch.valid.sum(0).float().mean().item()
        logs["final_turn"] = (
            (batch.turn * batch.valid).max(0).values.float().mean().item()
        )
        return logs

    def entropy(self):
        return


VTraceFromLogitsReturns = collections.namedtuple(
    "VTraceFromLogitsReturns",
    [
        "vs",
        "pg_advantages",
        "log_rhos",
        "behavior_action_log_probs",
        "target_action_log_probs",
        "kl_loss",
    ],
)

VTraceReturns = collections.namedtuple("VTraceReturns", "vs pg_advantages")


class Targets(NamedTuple):
    action_type_vtrace_returns: VTraceFromLogitsReturns = None
    move_vtrace_returns: VTraceFromLogitsReturns = None
    switch_vtrace_returns: VTraceFromLogitsReturns = None
    max_move_vtrace_returns: VTraceFromLogitsReturns = None
    flag_vtrace_returns: VTraceFromLogitsReturns = None
    target_vtrace_returns: VTraceFromLogitsReturns = None

    @property
    def batch_size(self):
        return self.action_type_vtrace_returns.vs.shape[1]

    @property
    def trajectory_length(self):
        return self.action_type_vtrace_returns.vs.shape[0]

    def iterate(
        self, minibatch_size: int = 16, minitraj_size: int = 16
    ) -> Iterator["Targets"]:
        for batch_index in range(math.ceil(self.batch_size / minibatch_size)):
            for traj_index in range(math.ceil(self.trajectory_length / minitraj_size)):
                batch_start = minibatch_size * batch_index
                batch_end = minibatch_size * (batch_index + 1)

                traj_start = minitraj_size * traj_index
                traj_end = minitraj_size * (traj_index + 1)

                minibatch = {}
                for key, lst in self._asdict().items():
                    if lst is not None:
                        value1, value2 = lst
                        minibatch[key] = [
                            value1[
                                traj_start:traj_end, batch_start:batch_end
                            ].contiguous(),
                            value2[
                                traj_start:traj_end, batch_start:batch_end
                            ].contiguous(),
                        ]

                yield Targets(**minibatch)

    def get_value_target(self, policy_field: str):
        return getattr(self, policy_field.replace("_policy", "_value_target"))

    def get_policy_target(self, policy_field: str):
        return getattr(self, policy_field + "_target")

    def get_has_played(self, policy_field: str, mask: torch.Tensor):
        has_played = [
            hp * mask
            for hp in getattr(self, policy_field.replace("_policy", "_has_played"))
        ]
        return has_played

    def slice(self, start: int, end: int, max_length: int = None) -> "Targets":
        targets_dict = {}
        for key, lst in self._asdict().items():
            if lst is not None:
                targets_dict[key] = VTraceFromLogitsReturns(
                    *[value[:max_length, start:end].contiguous() for value in lst]
                )
        return Targets(**targets_dict)



import os
import torch

from torch import nn

from typing import List

from meloetta.data import ROOT_DIR


class PokedexEmbedding(nn.Module):
    def __init__(self, gen: int, dtype: torch.dtype = torch.float32):
        super().__init__()

        embeddings: torch.Tensor
        names, embeddings = torch.load(f"{ROOT_DIR}/pretrained/gen{gen}/pokedex.pt")
        embeddings = embeddings.to(dtype)
        self.names: List[str] = names
        self.num_embeddings = embeddings.shape[0]
        self.embedding_dim = embeddings.shape[-1]
        self.emb = nn.Embedding.from_pretrained(embeddings)

    def get_name(self, x: torch.Tensor):
        return [self.names[token] for token in x.squeeze().tolist()]

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.emb(x)


class AbilityEmbedding(nn.Module):
    def __init__(self, gen: int, dtype: torch.dtype = torch.float32):
        super().__init__()

        embeddings: torch.Tensor
        names, embeddings = torch.load(f"{ROOT_DIR}/pretrained/gen{gen}/abilitydex.pt")
        embeddings = embeddings.to(dtype)
        self.names: List[str] = names
        self.num_embeddings = embeddings.shape[0]
        self.embedding_dim = embeddings.shape[-1]
        self.emb = nn.Embedding.from_pretrained(embeddings)

    def get_name(self, x: torch.Tensor):
        return [self.names[token] for token in x.squeeze().tolist()]

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.emb(x)


class MoveEmbedding(nn.Module):
    def __init__(self, gen: int, dtype: torch.dtype = torch.float32):
        super().__init__()

        embeddings: torch.Tensor
        names, embeddings = torch.load(f"{ROOT_DIR}/pretrained/gen{gen}/movedex.pt")
        embeddings = embeddings.to(dtype)
        self.names: List[str] = names
        self.num_embeddings = embeddings.shape[0]
        self.embedding_dim = embeddings.shape[-1]
        self.emb = nn.Embedding.from_pretrained(embeddings)

    def get_name(self, x: torch.Tensor):
        return [self.names[token] for token in x.squeeze().tolist()]

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.emb(x)


class ItemEmbedding(nn.Module):
    def __init__(self, gen: int, dtype: torch.dtype = torch.float32):
        super().__init__()

        embeddings: torch.Tensor
        names, embeddings = torch.load(f"{ROOT_DIR}/pretrained/gen{gen}/itemdex.pt")
        embeddings = embeddings.to(dtype)
        self.names: List[str] = names
        self.num_embeddings = embeddings.shape[0]
        self.embedding_dim = embeddings.shape[-1]
        self.emb = nn.Embedding.from_pretrained(embeddings)

    def get_name(self, x: torch.Tensor):
        return [self.names[token] for token in x.squeeze().tolist()]

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.emb(x)



import torch
import torch.nn as nn

from meloetta.frameworks.mewzero.model import config
from meloetta.frameworks.mewzero.model.utils import (
    _legal_policy,
    gather_along_rows,
)


class SwitchHead(nn.Module):
    def __init__(self, config: config.SwitchHeadConfig):
        super().__init__()

        self.key_fc = nn.Sequential(
            nn.Linear(config.entity_embedding_dim, config.entity_embedding_dim),
            nn.ReLU(),
            nn.Linear(config.entity_embedding_dim, config.key_dim),
        )
        self.query_fc = nn.Sequential(
            nn.Linear(config.autoregressive_embedding_dim, config.query_hidden_dim),
            nn.ReLU(),
            nn.Linear(config.query_hidden_dim, config.key_dim),
        )
        self.proj_switch = nn.Sequential(
            nn.Linear(config.key_dim, config.entity_embedding_dim),
            nn.ReLU(),
            nn.Linear(config.entity_embedding_dim, config.autoregressive_embedding_dim),
        )

    def forward(
        self,
        action_type_index: torch.Tensor,
        autoregressive_embedding: torch.Tensor,
        switches: torch.Tensor,
        switch_mask: torch.Tensor,
    ):
        T, B, *_ = autoregressive_embedding.shape

        keys = self.key_fc(switches)

        query = self.query_fc(autoregressive_embedding)
        query = query.view(T, B, 1, -1)

        switch_logits = query @ keys.transpose(-2, -1)
        switch_logits = switch_logits.view(T, B, -1)

        switch_mask = switch_mask.view(T * B, -1)
        switch_mask[switch_mask.sum() == 0] = True
        switch_mask = switch_mask.view(T, B, -1)

        switch_policy = _legal_policy(switch_logits, switch_mask)
        embedding_index = torch.multinomial(switch_policy.view(T * B, -1), 1)
        switch_index = embedding_index.view(T, B, -1)

        switch_embedding = gather_along_rows(keys.flatten(0, -3), embedding_index, 1)
        switch_embedding = switch_embedding.view(T, B, -1)
        projected_move_embedding = self.proj_switch(switch_embedding)

        valid_indices = action_type_index == 1
        valid_indices = valid_indices.unsqueeze(-1)

        return (
            switch_logits,
            switch_policy,
            switch_index,
            autoregressive_embedding + valid_indices * projected_move_embedding,
        )



import os
import re
import json

from extract import PS_CLIENT_DIR

CONDITIONS_PATH = f"pokemon-showdown-client/data/pokemon-showdown/data/conditions.ts"
MOVES_PATH = f"pokemon-showdown-client/data/pokemon-showdown/data/moves.ts"
BATTLE_PATH = f"pokemon-showdown-client/src/battle.ts"


def main():
    # os.system("npx prettier -w --tab-width 4 pokemon-showdown-client")

    src = ""
    with open(BATTLE_PATH, "r") as f:
        src += f.read()
    with open(CONDITIONS_PATH, "r") as f:
        src += f.read()
    with open(MOVES_PATH, "r") as f:
        src += f.read()

    for file in os.listdir(os.path.join(PS_CLIENT_DIR, "js")):
        if file.endswith(".js"):
            with open(
                os.path.join(PS_CLIENT_DIR, "js", file), "r", encoding="utf-8"
            ) as f:
                src += f.read()

    # with open("wsnc", "w", encoding="utf-8") as f:
    #     f.write(src)

    volatile_status = set()
    volatile_status.update(
        set(re.findall(r"removeVolatile\([\"|\'](.*?)[\"|\']\)", src))
    )
    volatile_status.update(set(re.findall(r"hasVolatile\([\"|\'](.*?)[\"|\']\)", src)))
    volatile_status.update(set(re.findall(r"volatiles\[[\"|\'](.*?)[\"|\']\]", src)))
    volatile_status.update(set(re.findall(r"volatiles\.(.*?)[\[|\)| ]", src)))
    volatile_status.update(
        set(re.findall(r"volatileStatus:\s*[\"|\'](.*)[\"|\'],", src))
    )
    volatile_status = list(volatile_status)
    for i, vs in enumerate(volatile_status):
        volatile_status[i] = "".join(c for c in vs if c.isalnum()).lower()
    volatile_status.sort()

    weathers = re.findall(r"[\"|\']-weather[\"|\'],\s*[\"|\'](.*)[\"|\'],", src)
    weathers = list(map(lambda s: s.lower(), sorted(set(weathers))))

    side_conditions = re.findall(r"sideCondition:\s*[\"|\'](.*)[\"|\'],", src)
    addSideCondition = re.search(
        r"addSideCondition\(.*\) \{([\S\s]*?)\n\t\}", src
    ).group()
    side_conditions += re.findall(r"case \"(.*?)\"", addSideCondition)
    side_conditions = list(sorted(set(side_conditions)))

    terrain = re.findall(r"terrain:\s*[\"|\'](.*)[\"|\'],", src)
    terrain = list(sorted(set(terrain)))

    pseudoweather = re.findall(r"pseudoWeather\:\s[\"|\'](.*?)[\"|\']", src)
    pseudoweather = list(sorted(set(pseudoweather)))

    item_effects = re.findall(r"itemEffect = [\"|\'](.*?)[\"|\']", src)
    item_effects = list(sorted(set(item_effects)))
    item_effects.pop(1)
    item_effects = set(item_effects)
    item_effects.update(["eaten", "popped", "consumed", "held up"])
    item_effects = list(sorted(item_effects))

    wsnc = {
        "weathers": weathers,
        "volatiles": volatile_status,
        "pseudoweather": pseudoweather,
        "terrain": terrain,
        "item_effects": item_effects,
        "side_conditions": side_conditions,
    }
    with open("meloetta/pretrained/wsnc.json", "w") as f:
        json.dump(wsnc, f)


if __name__ == "__main__":
    main()



import math
import torch
import torch.nn as nn
import torch.nn.functional as F

from typing import Optional


def binary_enc_matrix(num_embeddings: int):
    bits = math.ceil(math.log2(num_embeddings))
    mask = 2 ** torch.arange(bits)
    x = torch.arange(mask.sum().item() + 1)
    embs = x.unsqueeze(-1).bitwise_and(mask).ne(0).float()
    return embs[..., :num_embeddings:, :]


def sqrt_one_hot_matrix(num_embeddings: int):
    x = torch.arange(num_embeddings)
    i = (x**0.5).to(torch.long)
    return F.one_hot(i, torch.max(i) + 1).to(torch.float)


def _legal_policy(logits: torch.Tensor, legal_actions: torch.Tensor) -> torch.Tensor:
    """A soft-max policy that respects legal_actions."""
    # Fiddle a bit to make sure we don't generate NaNs or Inf in the middle.
    l_min = logits.min(axis=-1, keepdim=True).values
    logits = torch.where(legal_actions, logits, l_min)
    logits -= logits.max(axis=-1, keepdim=True).values
    logits *= legal_actions
    exp_logits = torch.where(
        legal_actions, torch.exp(logits), 0
    )  # Illegal actions become 0.
    exp_logits_sum = torch.sum(exp_logits, axis=-1, keepdim=True)
    policy = exp_logits / exp_logits_sum
    return policy


def _log_policy(logits: torch.Tensor, legal_actions: torch.Tensor) -> torch.Tensor:
    """Return the log of the policy on legal action, 0 on illegal action."""
    # logits_masked has illegal actions set to -inf.
    logits_masked = logits + torch.log(legal_actions)
    max_legal_logit = logits_masked.max(axis=-1, keepdim=True).values
    logits_masked = logits_masked - max_legal_logit
    # exp_logits_masked is 0 for illegal actions.
    exp_logits_masked = torch.exp(logits_masked)

    baseline = torch.log(torch.sum(exp_logits_masked, axis=-1, keepdim=True))
    # Subtract baseline from logits. We do not simply return
    #     logits_masked - baseline
    # because that has -inf for illegal actions, or
    #     legal_actions * (logits_masked - baseline)
    # because that leads to 0 * -inf == nan for illegal actions.
    log_policy = torch.multiply(legal_actions, (logits - max_legal_logit - baseline))
    return log_policy


def gather_along_rows(
    x: torch.Tensor, index: torch.Tensor, dim: int = 1, keepdim: bool = True
):
    """
    x: torch.Tensor -> [B, T, D]
    dim: int
    index: torch.Tensor -> [B,], values in range (0, T)
    """
    assert len(x.shape) == 3
    B, _, D = x.shape
    index = index.view(B, 1).repeat(1, D).view(B, 1, D)
    rows = torch.gather(x, dim, index)
    if keepdim:
        return rows
    else:
        return rows.squeeze(dim)


class GLU(nn.Module):
    def __init__(self, input_dim: int, output_dim: int, context_dim: int):
        super().__init__()
        self.lin1 = nn.Linear(context_dim, input_dim)
        self.lin2 = nn.Linear(input_dim, output_dim)

    def forward(self, x: torch.Tensor, context: torch.Tensor):
        gate = self.lin1(context)
        gate = torch.sigmoid(gate)
        x = gate * x
        x = self.lin2(x)
        return x


class Resblock(nn.Module):
    def __init__(
        self,
        hidden_size: int,
        num_layers: int = 2,
        use_layer_norm: bool = True,
    ) -> None:
        super().__init__()

        hiddens = [nn.ReLU(), nn.Linear(hidden_size, hidden_size)]
        if use_layer_norm:
            hiddens = [nn.LayerNorm(hidden_size)] + hiddens
        layer = lambda: nn.Sequential(*hiddens)
        self.layers = nn.ModuleList([layer() for _ in range(num_layers)])

    def forward(self, x: torch.Tensor):
        shortcut = x
        for layer in self.layers:
            x = layer(x)
        return x + shortcut


class TransformerLayer(nn.Module):
    def __init__(
        self,
        transformer_num_heads: int,
        transformer_key_size: int,
        transformer_value_size: int,
        use_layer_norm: bool = True,
    ) -> None:
        super().__init__()
        self._use_layer_norm = use_layer_norm
        if use_layer_norm:
            self.ln = nn.LayerNorm(transformer_key_size)
        self.mha = nn.MultiheadAttention(
            embed_dim=transformer_key_size,
            kdim=transformer_key_size,
            vdim=transformer_value_size,
            num_heads=transformer_num_heads,
            batch_first=True,
        )

    def forward(self, x: torch.Tensor, mask: torch.Tensor):
        x1 = x
        if self._use_layer_norm:
            x1 = self.ln(x1)

        x1 = F.relu(x1)
        # The logits mask has shape [num_heads, num_units, num_units]:
        x1, _ = self.mha(x1, x1, x1, key_padding_mask=mask)

        # Mask here mostly for safety:
        x1 = torch.where(~mask.unsqueeze(-1), x1, 0)
        x = x + x1

        return x


class TransformerEncoder(nn.Module):
    def __init__(
        self,
        num_layers: int,
        num_heads: int,
        key_size: int,
        value_size: int,
        resblocks_num_before: int,
        resblocks_num_after: int,
        resblocks_hidden_size: Optional[int] = None,
        use_layer_norm: bool = True,
    ):

        super().__init__()
        self._transformer_num_layers = num_layers
        self._transformer_num_heads = num_heads
        self._transformer_key_size = key_size
        self._transformer_value_size = value_size

        if resblocks_hidden_size is None:
            resblocks_hidden_size = key_size

        self._transformer_layers = nn.ModuleList(
            [
                TransformerLayer(
                    num_heads,
                    key_size,
                    value_size,
                    use_layer_norm,
                )
                for _ in range(num_layers)
            ]
        )

        self._resblocks_before = nn.ModuleList(
            [
                Resblock(
                    hidden_size=resblocks_hidden_size,
                    use_layer_norm=use_layer_norm,
                )
                for _ in range(resblocks_num_before)
            ]
        )
        self._resblocks_after = nn.ModuleList(
            [
                Resblock(
                    hidden_size=resblocks_hidden_size,
                    use_layer_norm=use_layer_norm,
                )
                for _ in range(resblocks_num_after)
            ]
        )

        self._resblocks_hidden_size = resblocks_hidden_size
        self._use_layer_norm = use_layer_norm

    def forward(self, x: torch.Tensor, mask: torch.Tensor = None) -> torch.Tensor:
        for resblock_before in self._resblocks_before:
            x = resblock_before(x)

        for transformer_layer in self._transformer_layers:
            x = transformer_layer(x, mask)

        for resblock_after in self._resblocks_after:
            x = resblock_after(x)

        return x


class AttentionPool(nn.Module):
    def __init__(self, embed_dim: int, nhead: int) -> None:
        super().__init__()

        self.nhead = nhead
        self.lin_qk = nn.Linear(embed_dim, 3 * embed_dim)
        self.proj = nn.Linear(embed_dim, embed_dim)

    def forward(self, x: torch.Tensor, mask: torch.Tensor):
        TB, S, C = x.shape

        q, k, v = torch.chunk(self.lin_qk(x), 3, 2)
        q = q.view(TB, S, self.nhead, C // self.nhead).transpose(1, 2)
        k = k.view(TB, S, self.nhead, C // self.nhead).transpose(1, 2)
        v = v.view(TB, S, self.nhead, C // self.nhead).transpose(1, 2)

        attn = (q @ k.transpose(-2, -1)) * (1.0 / math.sqrt(k.size(-1)))
        mask = (mask * mask.transpose(-2, -1)).unsqueeze(1)
        attn = torch.masked_fill(attn, mask, float("-inf"))
        score = F.softmax(attn, -1)

        y = score @ v
        y = y.transpose(1, 2).contiguous().view(TB, S, C)
        y = y.sum(1, keepdim=True)
        y = self.proj(y)

        return y


class ToVector(nn.Module):
    """Per-unit processing then average over the units dimension."""

    def __init__(self, input_dim: int, output_dim: int, hidden_size: int = 1024):
        super().__init__()
        self.ln1 = nn.LayerNorm(input_dim)
        self.ln2 = nn.LayerNorm(hidden_size)
        self.lin1 = nn.Linear(input_dim, hidden_size)
        self.lin2 = nn.Linear(hidden_size, output_dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.ln1(x)
        x = F.relu(x)
        x = self.lin1(x)
        x = torch.mean(x, 1)
        x = self.ln2(x)
        x = F.relu(x)
        x = self.lin2(x)
        return x



import torch
import torch.nn as nn

from typing import Tuple

from meloetta.frameworks.porygon.model import config
from meloetta.frameworks.porygon.model.interfaces import (
    EncoderOutput,
    Indices,
    Logits,
    Policy,
    State,
)
from meloetta.frameworks.porygon.model.heads import (
    ActionTypeHead,
    MoveHead,
    FlagsHead,
    SwitchHead,
    MaxMoveHead,
)
from meloetta.frameworks.porygon.model.utils import (
    _legal_policy,
    gather_along_rows,
)


class PolicyHeads(nn.Module):
    def __init__(self, gen: int, gametype: str, config: config.PolicyHeadsConfig):
        super().__init__()

        self.config = config
        self.gametype = gametype
        self.gen = gen

        self.action_type_head = ActionTypeHead(config.action_type_head_config)
        self.move_head = MoveHead(config.move_head_config)
        self.switch_head = SwitchHead(config.switch_head_config)

        self.proj_key = nn.Sequential(
            nn.Linear(config.key_dim, config.entity_embedding_dim),
            nn.ReLU(),
            nn.Linear(config.entity_embedding_dim, config.autoregressive_embedding_dim),
        )

        if gen == 8:
            self.max_move_head = MaxMoveHead(config.move_head_config)

        if gen >= 6:
            self.flag_head = FlagsHead(config.flag_head_config)

        # if gametype != "singles":
        #     self.target_head = PolicyHead(config.target_head_config)

    def forward(
        self,
        state_emb: torch.Tensor,
        encoder_output: EncoderOutput,
        state: State,
    ) -> Tuple[Indices, Logits, Policy]:

        T, B, *_ = state_emb.shape

        moves = encoder_output.side_embs.moves.squeeze(2)
        switches = encoder_output.side_embs.switches

        (
            action_type_logits,
            action_type_policy,
            action_type_index,
            at_autoregressive_embedding,
        ) = self.action_type_head(
            state_emb,
            state["action_type_mask"],
        )

        (
            move_logits,
            move_policy,
            move_index,
            autoregressive_embedding,
        ) = self.move_head(
            action_type_index,
            at_autoregressive_embedding,
            moves,
            state["move_mask"],
        )

        (
            switch_logits,
            switch_policy,
            switch_index,
            autoregressive_embedding,
        ) = self.switch_head(
            action_type_index,
            autoregressive_embedding,
            switches,
            state["switch_mask"],
        )

        if self.gen >= 6:
            (
                flag_logits,
                flag_policy,
                flag_index,
                autoregressive_embedding,
            ) = self.flag_head(
                action_type_index,
                autoregressive_embedding,
                state["flag_mask"],
            )
        else:
            flag_logits = None
            flag_policy = None
            flag_index = None

        if self.gen == 8:
            max_move_logits, max_move_policy, max_move_index = self.max_move_head(
                action_type_index,
                autoregressive_embedding,
                moves,
                state["max_move_mask"],
            )
        else:
            max_move_index = None
            max_move_logits = None
            max_move_policy = None

        if self.gametype != "singles":
            target_logits, target_policy, target_index = self.target_head(
                state["prev_choices"],
                moves,
                switches,
                action_type_index,
            )
        else:
            target_index = None
            target_logits = None
            target_policy = None

        indices = Indices(
            action_type_index=action_type_index,
            move_index=move_index,
            max_move_index=max_move_index,
            switch_index=switch_index,
            flag_index=flag_index,
            target_index=target_index,
        )
        logits = Logits(
            action_type_logits=action_type_logits,
            move_logits=move_logits,
            max_move_logits=max_move_logits,
            switch_logits=switch_logits,
            flag_logits=flag_logits,
            target_logits=target_logits,
        )
        policy = Policy(
            action_type_policy=action_type_policy,
            move_policy=move_policy,
            max_move_policy=max_move_policy,
            switch_policy=switch_policy,
            flag_policy=flag_policy,
            target_policy=target_policy,
        )

        return indices, logits, policy



import torch
import random
import threading
import multiprocessing as mp

from typing import Dict, List

from meloetta.frameworks.nash_ketchum.model.interfaces import Batch
from meloetta.frameworks.nash_ketchum.utils import create_buffers


Buffers = Dict[str, List[torch.Tensor]]


class ReplayBuffer:
    def __init__(
        self,
        trajectory_length: int,
        gen: int,
        gametype: str,
        num_buffers: int = 512,
        device: torch.device = torch.device(
            "cuda" if torch.cuda.is_available() else "cpu"
        ),
    ):
        self.num_buffers = num_buffers
        self.device = device
        self.finish_queue = mp.Queue()

        manager = mp.Manager()
        self.index_cache = manager.dict()

        self.buffers = create_buffers(num_buffers, trajectory_length, gen, gametype)

        self.valid_masks = [field for field in self.buffers if field.endswith("_mask")]

        self.done_cache = [0 for _ in range(num_buffers)]
        self.turn_counters = [0 for _ in range(num_buffers)]

        self.full_queue = mp.Queue()
        self.free_queue = mp.Queue()

        for m in range(num_buffers):
            self.free_queue.put(m)

    def _get_index(self, battle_tag: str) -> int:
        if battle_tag in self.index_cache:
            index = self.index_cache[battle_tag]
        else:
            index = self.free_queue.get()
            self._reset_index(index)
            self.index_cache[battle_tag] = index
            assert torch.all(~self.buffers["valid"][index]).item()
            assert self.turn_counters[index] == 0
        return index

    def store_sample(
        self,
        index: int,
        step: Dict[str, torch.Tensor],
        pid: int,
    ):
        turn = self.turn_counters[index]
        assert (~self.buffers["valid"][index][turn]).item()
        for key, value in step.items():
            try:
                self.buffers[key][index][turn][...] = value
            except Exception as e:
                raise e
        self.buffers["player_id"][index][turn][...] = pid
        self.buffers["valid"][index][turn][...] = 1
        self.turn_counters[index] += 1

    def append_reward(self, index: int, pid: int, reward: int):
        turn = self.turn_counters[index] - 1
        self.buffers["rewards"][index][turn, pid][...] = reward
        # self.buffers["valid"][index][turn][...] = 0

    def register_done(self, battle_tag: str):
        index = self._get_index(battle_tag)
        self.done_cache[index] += 1
        if self.done_cache[index] == 2:
            length = sum(self.buffers["valid"][index]).item()
            self.finish_queue.put((None, length))
            self._clear_cache(battle_tag, index)

    def _clear_cache(self, battle_tag: str, index: int):
        self.turn_counters[index] = 0
        self.full_queue.put(index)
        self.done_cache[index] = 0
        self.index_cache.pop(battle_tag)

    def _reset_index(self, index: int):
        self.buffers["valid"][index][...] = 0
        self.buffers["rewards"][index][...] = 0
        self.buffers["scalars"][index][...] = 0
        for valid_mask in self.valid_masks:
            self.buffers[valid_mask][index][...] = 1

    def get_batch(
        self,
        batch_size: int,
        lock=threading.Lock(),
    ) -> Batch:
        def _get_index():
            index = self.full_queue.get()
            return index

        with lock:
            indices = [_get_index() for _ in range(batch_size)]

        # for m in indices:
        #     end = self.buffers["valid"][m].sum(-1)
        #     assert self.buffers["valid"][m][:end].sum() == end
        #     assert self.buffers["valid"][m][end:].sum() == 0

        rewards = torch.stack([self.buffers["rewards"][m] for m in indices])
        if not (torch.sum(rewards, dim=-1) == 0).all().item():
            print("Batch rewards are not zero-sum!")

        valids = torch.stack([self.buffers["valid"][m] for m in indices])
        lengths = valids.sum(-1)  # + 1

        max_length = lengths.max().item()

        indices = list(
            zip(
                *sorted(
                    list(zip(lengths.tolist(), indices)),
                    key=lambda x: x[0],
                    reverse=True,
                )
            )
        )[1]

        batch = {
            key: torch.stack(
                [self.buffers[key][index][:max_length] for index in indices],
                dim=1,
            )
            for key in self.buffers
        }

        for m in indices:
            # if random.random() < 0.75:
            #     self.full_queue.put(m)
            # else:
            self.free_queue.put(m)

        return Batch(**batch)



import os
import time
import wandb
import cProfile, pstats
import traceback
import threading

import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F

from typing import Tuple, Dict

from meloetta.frameworks.nash_ketchum.buffer import ReplayBuffer
from meloetta.frameworks.nash_ketchum.config import NAshKetchumConfig
from meloetta.frameworks.nash_ketchum.entropy import EntropySchedule
from meloetta.frameworks.nash_ketchum.model import (
    NAshKetchumModel,
    ModelOutput,
    Batch,
    Targets,
    Loss,
    Indices,
)
from meloetta.frameworks.nash_ketchum.utils import (
    FineTuning,
    _player_others,
    v_trace,
    get_loss_nerd,
    get_loss_v,
)


class TargetNetSGD:
    def __init__(
        self,
        lr: float,
        target_model: NAshKetchumModel,
        learner_model: NAshKetchumModel,
    ):
        self.lr = lr
        self.target_model = target_model
        self.learner_model = learner_model

    @torch.no_grad()
    def step(self):
        for param1, param2 in zip(
            self.target_model.parameters(), self.learner_model.parameters()
        ):
            if param2.requires_grad:
                new = param1 + self.lr * (param2 - param1)
                param1[...] = new.clone().detach()


class NAshKetchumLearner:
    def __init__(
        self,
        learner_model: NAshKetchumModel = None,
        actor_model: NAshKetchumModel = None,
        target_model: NAshKetchumModel = None,
        model_prev: NAshKetchumModel = None,
        model_prev_: NAshKetchumModel = None,
        replay_buffer: ReplayBuffer = None,
        load_devices: bool = True,
        init_optimizers: bool = True,
        config: NAshKetchumConfig = NAshKetchumConfig(),
    ):
        self.config = config

        self.learner_model = (
            learner_model if learner_model is not None else self._init_model(train=True)
        )
        self.actor_model = (
            actor_model if actor_model is not None else self._init_model()
        )
        self.target_model = (
            target_model if target_model is not None else self._init_model()
        )
        self.model_prev = model_prev if model_prev is not None else self._init_model()
        self.model_prev_ = (
            model_prev_ if model_prev_ is not None else self._init_model()
        )

        self.actor_model.load_state_dict(self.learner_model.state_dict())
        self.target_model.load_state_dict(self.learner_model.state_dict())
        self.model_prev.load_state_dict(self.learner_model.state_dict())
        self.model_prev_.load_state_dict(self.learner_model.state_dict())

        self.actor_model.share_memory()

        if load_devices:
            self._to_devices()

        if init_optimizers:
            self._init_optimizers()

        self.replay_buffer = replay_buffer if replay_buffer else self._init_buffer()

        self._entropy_schedule = EntropySchedule(
            sizes=self.config.entropy_schedule_size,
            repeats=self.config.entropy_schedule_repeats,
        )
        self.scaler = torch.cuda.amp.GradScaler()

        self.learner_steps = 0
        self.finetune = FineTuning()

        print(f"learnabled params: {self.learner_model.get_learnable_params():,}")

    def get_config(self):
        return {
            "parameters": self.learner_model.get_learnable_params(),
            "learner_steps": self.learner_steps,
            **self.config.__dict__,
        }

    def save(self):
        torch.save(
            {
                "learner_model": self.learner_model.state_dict(),
                "actor_model": self.actor_model.state_dict(),
                "target_model": self.target_model.state_dict(),
                "model_prev": self.model_prev.state_dict(),
                "model_prev_": self.model_prev_.state_dict(),
                "config": self.config,
                "optimizer": self.optimizer.state_dict(),
                "learner_steps": self.learner_steps,
            },
            f"cpkts/cpkt-{self.learner_steps:08}.tar",
        )

    def _init_model(self, train: bool = False):
        model = NAshKetchumModel(
            gen=self.config.gen,
            gametype=self.config.gametype,
            config=self.config.model_config,
        ).float()
        if train:
            model.train()
        else:
            model.eval()
        return model

    def _init_buffer(self):
        return ReplayBuffer(
            self.config.trajectory_length,
            self.config.gen,
            self.config.gametype,
            self.config.num_buffers,
            self.config.learner_device,
        )

    def _init_optimizers(self):
        self.optimizer = optim.Adam(
            self.learner_model.parameters(),
            lr=self.config.learning_rate,
            betas=(self.config.adam.b1, self.config.adam.b2),
            eps=self.config.adam.eps,
        )
        self.optimizer_target = TargetNetSGD(
            self.config.target_network_avg,
            self.target_model,
            self.learner_model,
        )

    def _to_devices(self):
        self.learner_model = self.learner_model.to(self.config.learner_device)
        self.actor_model = self.actor_model.to(self.config.actor_device)
        self.target_model = self.target_model.to(self.config.learner_device)
        self.model_prev = self.model_prev.to(self.config.learner_device)
        self.model_prev_ = self.model_prev_.to(self.config.learner_device)

    @classmethod
    def from_config(self, config: NAshKetchumConfig = None):
        return NAshKetchumLearner.from_fpath(config_override=config)

    @classmethod
    def from_latest(self):
        cpkts = list(
            sorted(
                os.listdir("cpkts"),
                key=lambda f: int(f.split("-")[-1].split(".")[0]),
            )
        )
        fpath = os.path.join("cpkts", cpkts[-1])
        return self.from_fpath(fpath, NAshKetchumConfig())

    @classmethod
    def from_fpath(self, fpath: str, config: NAshKetchumConfig = None):
        try:
            print(f"Using fpath: {fpath}")
            datum: dict = torch.load(fpath)
        except Exception as e:
            raise ValueError(f"cpkt at {fpath} not found")
        else:
            if config:
                print("Using new config!")
            else:
                print("Found config!")
                config: NAshKetchumConfig = datum["config"]

        learner = NAshKetchumLearner(
            config=config, load_devices=False, init_optimizers=False
        )

        for obj, model in [
            (learner.learner_model, "learner_model"),
            (learner.actor_model, "actor_model"),
            (learner.target_model, "target_model"),
            (learner.model_prev, "model_prev"),
            (learner.model_prev_, "model_prev_"),
        ]:
            try:
                obj.load_state_dict(datum[model], strict=False)
            except Exception as e:
                print(f"Error loading `{model}`")
                traceback.print_exc()
            else:
                print(f"Sucessfully loaded `{model}`")

        learner._to_devices()
        learner._init_optimizers()

        try:
            learner.optimizer.load_state_dict(datum["optimizer"])
        except Exception as e:
            print("Optimizer not loaded")
            traceback.print_exc()
        else:
            print(f"Optimizer loaded sucessfully!")

        if datum.get("learner_steps"):
            learner.learner_steps = datum["learner_steps"]
        else:
            print("Learner steps not loaded")

        return learner

    def collect_batch_trajectory(self):
        return self.replay_buffer.get_batch(self.config.batch_size)

    def step(self):
        batch = self.collect_batch_trajectory()
        alpha, update_target_net = self._entropy_schedule(self.learner_steps)

        if self.learner_steps < int(1e9):
            # if self.learner_steps < 10:
            self.update_paramters(batch, alpha, update_target_net)
        else:
            profiler = cProfile.Profile()
            profiler.enable()
            self.update_paramters(batch, alpha, update_target_net)
            profiler.disable()
            stats = pstats.Stats(profiler).sort_stats("ncalls")
            fpath = f"cprofile_{self.learner_steps}"
            stats.dump_stats(fpath)

    def update_paramters(
        self,
        batch: Batch,
        alpha: float,
        indices: Indices,
    ):
        batch_cpu = batch
        batch_gpu = batch.to(self.replay_buffer.device)

        indices = Indices(
            action_type_index=getattr(batch_cpu, "action_type_index", None),
            move_index=getattr(batch_cpu, "move_index", None),
            max_move_index=getattr(batch_cpu, "max_move_index", None),
            switch_index=getattr(batch_cpu, "switch_index", None),
            flag_index=getattr(batch_cpu, "flag_index", None),
            target_index=getattr(batch_cpu, "target_index", None),
        )

        targets = self._get_targets(batch_cpu, batch_gpu, alpha, indices)
        self.backprop(batch_cpu, targets, indices)

    def backprop(
        self,
        batch_cpu: Batch,
        targets: Targets,
        indices: Indices,
        lock=threading.Lock(),
    ):
        _, update_target_net = self._entropy_schedule(self.learner_steps)

        losses = self._backpropagate(batch_cpu, targets, indices)

        static_loss_dict = losses._asdict()
        static_loss_dict["s"] = self.learner_steps
        static_loss_dict["final_turn"] = (
            batch_cpu.scalars[..., 0].max(0).values.float().mean()
        ).item()

        wandb.log(static_loss_dict)

        self.scaler.unscale_(self.optimizer)

        torch.nn.utils.clip_grad_value_(
            self.learner_model.parameters(), self.config.clip_gradient
        )

        self.scaler.step(self.optimizer)
        self.scaler.update()
        self.optimizer.zero_grad()

        self.optimizer_target.step()

        if update_target_net:
            print(f"Updating Regularization Policy @ Step: {self.learner_steps:,}")
            self.model_prev_.load_state_dict(self.model_prev.state_dict())
            self.model_prev.load_state_dict(self.target_model.state_dict())

        self.learner_steps += 1
        if self.learner_steps % 1000 == 0:
            self.save()

        with lock:
            self.actor_model.load_state_dict(self.learner_model.state_dict())

    @torch.no_grad()
    def _learning_forward(
        self, batch: Batch, indices: Indices, size: int = 4096
    ) -> Tuple[ModelOutput, ModelOutput, ModelOutput, ModelOutput,]:

        B, T = batch.batch_size, batch.trajectory_length
        if B * T >= size:

            learner = []
            target = []
            prev = []
            prev_ = []

            def _iterate(
                batch: Dict[str, torch.Tensor],
                indices: Dict[str, torch.Tensor],
                size: int,
            ):
                n = batch["valid"].shape[1] // size
                if batch["valid"].shape[1] / size != n:
                    n += 1

                for i in range(n):
                    start, end = i * size, (i + 1) * size
                    yield (
                        {k: v[:, start:end] for k, v in batch.items()},
                        {k: v[:, start:end] for k, v in indices.items()},
                    )

            for minibatch, minindex in _iterate(
                batch.flatten(0, 1), indices.flatten(0, 1), size
            ):
                minindex = Indices(**minindex)
                learner.append(self.learner_model.learning_forward(minibatch, minindex))
                target.append(self.target_model.learning_forward(minibatch, minindex))
                prev.append(self.model_prev.learning_forward(minibatch, minindex))
                prev_.append(self.model_prev_.learning_forward(minibatch, minindex))

            learner = ModelOutput.from_list(learner)
            target = ModelOutput.from_list(target)
            prev = ModelOutput.from_list(prev)
            prev_ = ModelOutput.from_list(prev_)

            learner = learner.view(T, B)
            target = target.view(T, B)
            prev = prev.view(T, B)
            prev_ = prev_.view(T, B)

        else:
            batch = batch._asdict()
            learner = self.learner_model.learning_forward(batch, indices)
            target = self.target_model.learning_forward(batch, indices)
            prev = self.model_prev.learning_forward(batch, indices)
            prev_ = self.model_prev_.learning_forward(batch, indices)

        return learner, target, prev, prev_

    def _get_targets(
        self, batch_cpu: Batch, batch_gpu: Batch, alpha: float, indices: Indices
    ) -> Targets:
        learner, target, prev, prev_ = self._learning_forward(
            batch_gpu, indices.to(self.replay_buffer.device)
        )

        learner = learner.to("cpu")
        target = target.to("cpu")
        prev = prev.to("cpu")
        prev_ = prev_.to("cpu")

        targets_dict = {"importance_sampling_corrections": []}

        acting_policies = []
        policies_valid = []
        policies_pprocessed = []
        log_policies_reg = []
        action_ohs = []
        fields = []

        for field in sorted(learner.pi._fields):

            pi = getattr(learner.pi, field)
            if pi is None:
                continue

            fields.append(field)

            log_pi = getattr(target.log_pi, field)
            prev_log_pi = getattr(prev.log_pi, field)
            prev_log_pi_ = getattr(prev_.log_pi, field)

            # mask_field = policy_field.replace("_policy", "_mask")
            # policy_mask = getattr(batch_cpu, mask_field)
            # policy_pprocessed = self.finetune(pi, policy_mask, self.learner_steps)

            if field == "move" or field == "flag":
                policy_valid = batch_cpu.valid & (batch_cpu.action_type_index == 0)

            elif field == "switch":
                policy_valid = batch_cpu.valid & (batch_cpu.action_type_index == 1)

            else:
                policy_valid = batch_cpu.valid
            policy_valid = policy_valid.clone()

            policy_pprocessed = pi
            acting_policy = getattr(batch_cpu, field + "_policy")

            log_policy_reg = log_pi - (alpha * prev_log_pi + (1 - alpha) * prev_log_pi_)
            # log_policy_reg = log_policy_reg * policy_valid.unsqueeze(-1)

            action_index = getattr(batch_cpu, field + "_index")
            action_oh = F.one_hot(action_index, pi.shape[-1])

            policies_valid.append(policy_valid)
            acting_policies.append(acting_policy)
            policies_pprocessed.append(policy_pprocessed)
            log_policies_reg.append(log_policy_reg)
            action_ohs.append(action_oh)

        (
            v_target_list,
            has_played_list,
            v_trace_policy_target_list,
            importance_sampling_corrections,
        ) = ([], [], [], [])

        for player in range(2):
            reward = batch_cpu.rewards[:, :, player]  # [T, B, Player]

            v_target_, has_played, policy_targets_, policy_ratios = v_trace(
                target.v,
                batch_cpu.valid,
                policies_valid,
                batch_cpu.player_id,
                acting_policies,
                policies_pprocessed,
                log_policies_reg,
                _player_others(batch_cpu.player_id, batch_cpu.valid, player),
                action_ohs,
                reward,
                player,
                lambda_=self.config.lambda_vtrace,
                c=self.config.c_vtrace,
                rho=self.config.rho_vtrace,
                eta=self.config.eta_reward_transform,
                gamma=self.config.gamma,
            )
            v_target_list.append(v_target_)
            has_played_list.append(has_played)
            v_trace_policy_target_list.append(
                {
                    key: policy_target_
                    for key, policy_target_ in zip(fields, policy_targets_)
                }
            )
            importance_sampling_corrections.append(
                {
                    key: policy_ratio.unsqueeze(-1)
                    for key, policy_ratio in zip(fields, policy_ratios)
                }
            )

        targets_dict["value_targets"] = v_target_list
        targets_dict["has_played"] = has_played_list

        for field in fields:
            targets_dict[f"{field}_policy_target"] = [
                targ[field] for targ in v_trace_policy_target_list if field in targ
            ]
            targets_dict[f"{field}_is"] = [
                targ[field] for targ in importance_sampling_corrections if field in targ
            ]

        return Targets(**{key: value for key, value in targets_dict.items() if value})

    def _backpropagate(
        self,
        batch: Batch,
        targets: Targets,
        indices: Indices,
        remove_padding: bool = True,
    ) -> Loss:

        loss_dict = {
            key + "_loss": 0
            for key in batch._fields
            if "_policy" in key and getattr(batch, key) is not None
        }
        loss_dict["value_loss"] = 0

        batch_size = batch.batch_size * batch.trajectory_length

        if remove_padding:
            t = batch.valid.sum(0)
            # old_size = batch.trajectory_length * batch.batch_size
            # reduced_size = batch.valid.sum().item()
            # reduction = 100 * (1 - (reduced_size / old_size))
            # print(
            #     f"\n{reduction:.2f} % reduction - bs_eff = {old_size}, true_bs = {reduced_size}"
            # )

            batch = batch.flatten_without_padding(t)
            targets = targets.flatten_without_padding(t)
            indices = indices.flatten_without_padding(t)

        else:
            batch = batch.flatten(0, 1)
            targets = targets.flatten(0, 1)
            indices = indices.flatten(0, 1)

        batch = {k: v.to(self.replay_buffer.device) for k, v in batch.items()}
        indices = {k: v.to(self.replay_buffer.device) for k, v in indices.items()}
        targets = {
            k: [v.to(self.replay_buffer.device) for v in vt]
            for k, vt in targets.items()
        }

        minibatch_size = 4096

        value_loss_field = "value_loss"

        n = batch_size // minibatch_size
        if batch["valid"].shape[1] / minibatch_size != n:
            n += 1

        for i in range(n):

            start, end = i * minibatch_size, (i + 1) * minibatch_size

            minibatch = {k: v[:, start:end] for k, v in batch.items()}

            if torch.all(~minibatch["valid"]):
                continue

            loss = 0
            minitarget = {k: [v[:, start:end] for v in vt] for k, vt in targets.items()}
            minindices = Indices(**{k: vt[:, start:end] for k, vt in indices.items()})

            with torch.autocast(device_type="cuda", dtype=torch.float16):
                model_output = self.learner_model.forward(minibatch, indices=minindices)

            for pi_field in model_output.pi._fields:

                field = pi_field.replace("_policy", "")
                pi = getattr(model_output.pi, pi_field)

                if pi is None:
                    continue

                mask_field = field + "_mask"
                logit = getattr(model_output.logit, field)

                if field == "move" or field == "flag":
                    valid = minibatch["valid"] * (minibatch["action_type_index"] == 0)
                    normalization = batch["valid"] * (batch["action_type_index"] == 0)

                elif field == "switch":
                    valid = minibatch["valid"] * (minibatch["action_type_index"] == 1)
                    normalization = batch["valid"] * (batch["action_type_index"] == 1)

                else:
                    valid = minibatch["valid"]
                    normalization = batch["valid"]

                policy_mask = minibatch[mask_field]

                policy_loss_field = field + "_policy_loss"

                # Uses v-trace to define q-values for Nerd
                policy_loss = get_loss_nerd(
                    [logit] * 2,
                    [pi] * 2,
                    minitarget[f"{field}_policy_target"],
                    valid,
                    minibatch["player_id"],
                    policy_mask,
                    minitarget[f"{field}_is"],
                    normalization=normalization,
                    clip=self.config.nerd.clip,
                    threshold=self.config.nerd.beta,
                )
                loss_dict[policy_loss_field] += policy_loss.item()

                loss += policy_loss

            # repr_loss = self._get_repr_loss(
            #     model_output.state_action_emb,
            #     model_output.state_emb,
            #     (batch["valid"][:-1] * batch["valid"][1:]).flatten(0, 1),
            # )
            # loss_dict["repr_loss"] = repr_loss.item()
            # loss = loss + repr_loss

            has_played = minitarget["has_played"]
            value_loss = get_loss_v(
                [model_output.v] * 2,
                minitarget["value_targets"],
                has_played,
                normalization=targets["has_played"],
            )
            loss_dict[value_loss_field] += value_loss.item()

            loss += value_loss

            self.scaler.scale(loss).backward()

        return Loss(**loss_dict)

    def _get_repr_loss(
        self,
        state_action_emb: torch.Tensor,
        state_emb: torch.Tensor,
        valid: torch.Tensor,
    ) -> torch.Tensor:
        criterion = nn.CrossEntropyLoss(reduction="none")

        n = state_action_emb[:-1].flatten(0, 1).shape[0]
        labels = torch.arange(n, device=self.replay_buffer.device)

        logits = torch.einsum(
            "ik,jk->ij",
            state_action_emb[:-1].flatten(0, 1),
            state_emb[1:].flatten(0, 1),
        )
        loss = criterion(logits, labels)
        loss = torch.maximum(loss, torch.tensor(1))

        loss = loss * valid
        loss = loss.sum() / valid.sum().clamp(min=1)

        return loss

    def run(self):
        while True:
            self.step()



import torch
import torch.nn as nn
import torch.nn.functional as F

from meloetta.frameworks.nash_ketchum.model import config
from meloetta.frameworks.nash_ketchum.model.utils import sqrt_one_hot_matrix, MLP

from meloetta.data import CHOICE_FLAGS, CHOICE_TARGETS, CHOICE_TOKENS

DRAW_BY_TURNS = 300


class ScalarEncoder(nn.Module):
    def __init__(self, gen: int, n_active: int, config: config.ScalarEncoderConfig):
        super().__init__()

        self.gen = gen
        self.turn1_sqrt_onehot = nn.Embedding.from_pretrained(
            sqrt_one_hot_matrix(DRAW_BY_TURNS + 2)
        )

        action_mask_size = (
            3  # action types
            + 4  # move indices
            + 6  # switch indices
            + 5  # move flags
        )
        if gen == 8:
            action_mask_size += 4  # max move indices

        if n_active > 1:
            action_mask_size += 2 * n_active  # n_targets

        self.n_onehot = nn.Embedding.from_pretrained(torch.eye(7))

        lin_in = (
            self.turn1_sqrt_onehot.embedding_dim
            + action_mask_size
            + 1
            + self.n_onehot.embedding_dim * 2 * 3
        )
        self.n_active = n_active

        if n_active > 1:
            prev_choice_token_onehot = nn.Embedding.from_pretrained(
                torch.eye(len(CHOICE_TOKENS) + 1)
            )
            self.prev_choice_token_onehot = nn.Sequential(
                prev_choice_token_onehot,
                nn.Flatten(2),
            )
            prev_choice_index_onehot = nn.Embedding.from_pretrained(torch.eye(7))
            self.prev_choice_index_onehot = nn.Sequential(
                prev_choice_index_onehot,
                nn.Flatten(2),
            )
            prev_choice_flag_token_onehot = nn.Embedding.from_pretrained(
                torch.eye(len(CHOICE_FLAGS) + 1)
            )
            self.prev_choice_flag_token_onehot = nn.Sequential(
                prev_choice_flag_token_onehot,
                nn.Flatten(2),
            )
            prev_choice_targ_token_onehot = nn.Embedding.from_pretrained(
                torch.eye(len(CHOICE_TARGETS) + 1)
            )
            self.prev_choice_targ_token_onehot = nn.Sequential(
                prev_choice_targ_token_onehot,
                nn.Flatten(2),
            )
            lin_in += n_active * (
                prev_choice_token_onehot.embedding_dim
                + prev_choice_index_onehot.embedding_dim
                + prev_choice_flag_token_onehot.embedding_dim
                + prev_choice_targ_token_onehot.embedding_dim
            )
            self.choices_done_onehot = nn.Embedding.from_pretrained(
                torch.eye(n_active * 2)
            )
            lin_in += self.choices_done_onehot.embedding_dim

        self.encoder = MLP([lin_in, config.embedding_dim, config.embedding_dim])

    def forward(
        self,
        turn: torch.Tensor,
        action_type_mask: torch.Tensor,
        move_mask: torch.Tensor,
        max_move_mask: torch.Tensor,
        switch_mask: torch.Tensor,
        flag_mask: torch.Tensor,
        target_mask: torch.Tensor,
        n: torch.Tensor,
        total_pokemon: torch.Tensor,
        faint_counter: torch.Tensor,
    ):
        turn = turn.clamp(min=0, max=DRAW_BY_TURNS)
        scalar_emb = [
            self.turn1_sqrt_onehot(turn),
            (turn / DRAW_BY_TURNS).unsqueeze(-1),
            action_type_mask,
            move_mask,
            switch_mask,
            flag_mask,
            self.n_onehot(n).flatten(-2),
            self.n_onehot(total_pokemon).flatten(-2),
            self.n_onehot(faint_counter).flatten(-2),
        ]
        if self.gen == 8:
            scalar_emb.append(max_move_mask)

        if self.n_active > 1:
            prev_choices_x = prev_choices + 1
            prev_choice_token = prev_choices_x[..., 0]
            prev_choice_index = prev_choices[..., 1] % 4
            prev_choice_index = prev_choice_index + 1
            prev_choice_index[prev_choices[..., 1] == -1] = 0
            prev_choice_flag_token = prev_choices_x[..., 2]
            prev_choice_targ_token = prev_choices_x[..., 3]
            scalar_emb += [
                self.choices_done_onehot(choices_done),
                self.prev_choice_token_onehot(prev_choice_token),
                self.prev_choice_index_onehot(prev_choice_index),
                self.prev_choice_flag_token_onehot(prev_choice_flag_token),
                self.prev_choice_targ_token_onehot(prev_choice_targ_token),
                target_mask,
            ]

        scalar_raw = torch.cat(scalar_emb, dim=-1)
        return self.encoder(scalar_raw)



import torch

from torch import nn
from torch.nn import functional as F

from meloetta.frameworks.mewzero.model import config
from meloetta.frameworks.mewzero.model.utils import (
    sqrt_one_hot_matrix,
    GLU,
    TransformerEncoder,
    ToVector,
)
from meloetta.frameworks.mewzero.model.encoders import private_spatial
from meloetta.embeddings import (
    AbilityEmbedding,
    PokedexEmbedding,
    MoveEmbedding,
    ItemEmbedding,
)

from meloetta.data import (
    GENDERS,
    STATUS,
    TOKENIZED_SCHEMA,
    BattleTypeChart,
)


class PrivateEncoder(nn.Module):
    def __init__(self, gen: int, n_active: int, config: config.PrivateEncoderConfig):
        super().__init__()

        self.config = config
        self.gen = gen
        self.n_active = n_active

        # onehots
        self.active_onehot = nn.Embedding.from_pretrained(torch.eye(3))
        self.fainted_onehot = nn.Embedding.from_pretrained(torch.eye(3))
        self.gender_onehot = nn.Embedding.from_pretrained(torch.eye(len(GENDERS) + 1))
        self.status_onehot = nn.Embedding.from_pretrained(torch.eye(len(STATUS) + 1))
        self.forme_embedding = nn.Embedding.from_pretrained(
            torch.eye(len(TOKENIZED_SCHEMA[f"gen{gen}"]["pokedex"]["forme"]) + 1)
        )

        self.hp_sqrt_onehot = nn.Embedding.from_pretrained(sqrt_one_hot_matrix(2048))
        self.level_sqrt_onehot = nn.Embedding.from_pretrained(sqrt_one_hot_matrix(102))

        self.atk_sqrt_onehot = nn.Embedding.from_pretrained(sqrt_one_hot_matrix(1024))
        self.def_sqrt_onehot = nn.Embedding.from_pretrained(sqrt_one_hot_matrix(1024))
        self.spa_sqrt_onehot = nn.Embedding.from_pretrained(sqrt_one_hot_matrix(1024))
        self.spd_sqrt_onehot = nn.Embedding.from_pretrained(sqrt_one_hot_matrix(1024))
        self.spe_sqrt_onehot = nn.Embedding.from_pretrained(sqrt_one_hot_matrix(1024))

        self.pp_sqrt_onehot = nn.Embedding.from_pretrained(sqrt_one_hot_matrix(64))

        # precomputed embeddings
        self.ability_embedding = AbilityEmbedding(gen=gen)
        self.pokedex_embedding = PokedexEmbedding(gen=gen)
        self.move_embedding = MoveEmbedding(gen=gen)
        self.item_embedding = ItemEmbedding(gen=gen)

        hp_embedding_dim = self.hp_sqrt_onehot.embedding_dim + 1
        level_embedding_dim = self.level_sqrt_onehot.embedding_dim + 1
        stat_embedding_dim = (
            self.atk_sqrt_onehot.embedding_dim
            + self.def_sqrt_onehot.embedding_dim
            + self.spa_sqrt_onehot.embedding_dim
            + self.spd_sqrt_onehot.embedding_dim
            + self.spe_sqrt_onehot.embedding_dim
        )
        move_embedding_dim = (
            self.move_embedding.embedding_dim + self.pp_sqrt_onehot.embedding_dim
        )
        entity_in = (
            self.ability_embedding.embedding_dim
            + self.pokedex_embedding.embedding_dim
            + self.item_embedding.embedding_dim
            + self.active_onehot.embedding_dim
            + self.fainted_onehot.embedding_dim
            + self.gender_onehot.embedding_dim
            + hp_embedding_dim
            + level_embedding_dim
            + self.forme_embedding.embedding_dim
            + stat_embedding_dim
            + self.status_onehot.embedding_dim
        )
        if gen == 9:
            self.commanding_onehot = nn.Embedding.from_pretrained(torch.eye(3))
            self.reviving_onehot = nn.Embedding.from_pretrained(torch.eye(3))
            self.tera_onehot = nn.Embedding.from_pretrained(torch.eye(3))
            self.teratype_onehot = nn.Embedding.from_pretrained(
                torch.eye(len(BattleTypeChart) + 1)
            )
            entity_in += (
                self.commanding_onehot.embedding_dim
                + self.reviving_onehot.embedding_dim
                + self.tera_onehot.embedding_dim
                + self.teratype_onehot.embedding_dim
            )

        elif gen == 8:
            self.can_gmax_onehot = nn.Embedding.from_pretrained(torch.eye(3))
            entity_in += self.can_gmax_onehot.embedding_dim

        self.move_lin = nn.Linear(move_embedding_dim, config.entity_embedding_dim)
        self.entity_lin = nn.Linear(entity_in, config.entity_embedding_dim)

        self.transformer = TransformerEncoder(
            key_size=config.entity_embedding_dim,
            value_size=config.entity_embedding_dim,
            num_heads=config.transformer_num_heads,
            num_layers=config.transformer_num_layers,
            resblocks_num_before=config.resblocks_num_before,
            resblocks_num_after=config.resblocks_num_after,
        )
        self.output = ToVector(
            input_dim=config.entity_embedding_dim,
            output_dim=2 * config.entity_embedding_dim,
        )
        self.moves_w_context = GLU(
            config.entity_embedding_dim,
            config.entity_embedding_dim,
            config.entity_embedding_dim,
        )
        self.spatial = private_spatial.PrivateSpatialEncoder(
            gen=gen, n_active=n_active, config=config
        )

    def embed_hp(self, hp: torch.Tensor, maxhp: torch.Tensor) -> torch.Tensor:
        hp_ratio = (hp / maxhp.clamp(min=1)).unsqueeze(-1)
        hp_sqrt_onehot = self.hp_sqrt_onehot(hp)
        hp_sqrt_onehot = torch.cat([hp_ratio, hp_sqrt_onehot], dim=-1)
        return hp_sqrt_onehot

    def embed_stat(
        self,
        stat_atk: torch.Tensor,
        stat_def: torch.Tensor,
        stat_spa: torch.Tensor,
        stat_spd: torch.Tensor,
        stat_spe: torch.Tensor,
    ) -> torch.Tensor:
        stat_atk_emb = self.atk_sqrt_onehot(stat_atk)
        stat_def_emb = self.def_sqrt_onehot(stat_def)
        stat_spa_emb = self.spa_sqrt_onehot(stat_spa)
        stat_spd_emb = self.spd_sqrt_onehot(stat_spd)
        stat_spe_emb = self.spe_sqrt_onehot(stat_spe)
        stat_emb = torch.cat(
            [
                stat_atk_emb,
                stat_def_emb,
                stat_spa_emb,
                stat_spd_emb,
                stat_spe_emb,
            ],
            dim=-1,
        )
        return stat_emb

    def embed_level(self, level: torch.Tensor) -> torch.Tensor:
        level_sqrt_onehot = self.level_sqrt_onehot(level)
        level_ratio = (level / 100).unsqueeze(-1)
        level_emb = torch.cat((level_sqrt_onehot, level_ratio), dim=-1)
        return level_emb

    def embed_pp(self, pp: torch.Tensor) -> torch.Tensor:
        pp_bin = self.pp_sqrt_onehot(pp)
        return pp_bin

    def embed_moves(
        self, move_tokens: torch.Tensor, move_used: torch.Tensor
    ) -> torch.Tensor:
        moves_emb = self.move_embedding(move_tokens)
        move_used_emb = self.embed_pp(move_used)
        moves_emb = torch.cat((moves_emb, move_used_emb), dim=-1)
        return self.move_lin(moves_emb)

    def moves_given_context(self, active: torch.Tensor, moves: torch.Tensor):
        active = active.unsqueeze(-2).repeat_interleave(4, -2)
        moves = moves[..., : self.n_active, :, :]
        return self.moves_w_context(moves, active)

    def forward(self, private_reserve: torch.Tensor):
        private_reserve_x = private_reserve + 1
        ability = private_reserve_x[..., 0]
        active = private_reserve_x[..., 1]
        fainted = private_reserve_x[..., 2]
        gender = private_reserve_x[..., 3]
        hp = private_reserve_x[..., 4]
        item = private_reserve_x[..., 5]
        level = private_reserve_x[..., 6]
        maxhp = private_reserve_x[..., 7]
        name = private_reserve_x[..., 8]
        forme = private_reserve_x[..., 9]
        stat_atk = private_reserve_x[..., 10]
        stat_def = private_reserve_x[..., 11]
        stat_spa = private_reserve_x[..., 12]
        stat_spd = private_reserve_x[..., 13]
        stat_spe = private_reserve_x[..., 14]
        status = private_reserve_x[..., 15]

        if self.gen == 9:
            commanding = private_reserve_x[..., 16]
            reviving = private_reserve_x[..., 17]
            teraType = private_reserve_x[..., 18]
            terastallized = private_reserve_x[..., 19]

        elif self.gen == 8:
            canGmax = private_reserve_x[..., 16]

        moves = private_reserve_x[..., -8:]
        moves = moves.view(*moves.shape[:-1], 4, 2)
        move_tokens = moves[..., 0]
        move_used = moves[..., 1]

        ability_emb = self.ability_embedding(ability)
        active_emb = self.active_onehot(active)
        fainted_emb = self.fainted_onehot(fainted)
        gender_emb = self.gender_onehot(gender)
        hp_emb = self.embed_hp(hp, maxhp)
        item_emb = self.item_embedding(item)
        level_emb = self.embed_level(level)
        name_emb = self.pokedex_embedding(name)
        forme_emb = self.forme_embedding(forme)
        stat_emb = self.embed_stat(stat_atk, stat_def, stat_spa, stat_spd, stat_spe)
        status_emb = self.status_onehot(status)
        moves_emb = self.embed_moves(move_tokens, move_used)

        mon_emb = [
            ability_emb,
            active_emb,
            fainted_emb,
            gender_emb,
            hp_emb,
            item_emb,
            level_emb,
            name_emb,
            forme_emb,
            stat_emb,
            status_emb,
        ]

        if self.gen == 9:
            commanding_onehot = self.commanding_onehot(commanding)
            reviving_onehot = self.reviving_onehot(reviving)
            teraType_onehot = self.teratype_onehot(teraType)
            terastallized_onehot = self.tera_onehot(terastallized)
            mon_emb += [
                commanding_onehot,
                reviving_onehot,
                teraType_onehot,
                terastallized_onehot,
            ]

        elif self.gen == 8:
            can_gmax_emb = self.can_gmax_onehot(canGmax)
            mon_emb += [can_gmax_emb]

        entity_embeddings = torch.cat(mon_emb, dim=-1)
        entity_embeddings = self.entity_lin(entity_embeddings)
        entity_embeddings = entity_embeddings + moves_emb.sum(-2)

        T, B, S, *_ = entity_embeddings.shape

        mask = (name == 0) | (fainted == 2)
        mask = mask.bool()
        mask = mask.view(T * B, S)

        encoder_mask = mask.clone()
        encoder_mask[..., 0] = False

        entity_embeddings: torch.Tensor = entity_embeddings.view(T * B, S, -1)
        entity_embeddings = self.transformer(entity_embeddings, encoder_mask)

        entity_embeddings = entity_embeddings.view(T * B, S, -1)
        mask = mask.view(T * B, S, -1)

        entity_embeddings = entity_embeddings * ~mask
        entity_embedding = self.output(entity_embeddings)
        entity_embedding = entity_embedding.view(T, B, -1)

        spatial_embedding = self.spatial(entity_embeddings, ~mask)
        spatial_embedding = spatial_embedding.view(T, B, -1)

        entity_embeddings = entity_embeddings.view(T, B, S, -1)

        active = entity_embeddings[..., : self.n_active, :]
        moves = self.moves_given_context(active, moves_emb)

        return entity_embedding, spatial_embedding, moves, entity_embeddings


class PrivateEncoderV2(nn.Module):
    def __init__(self, gen: int, n_active: int, config: config.PrivateEncoderConfig):
        super().__init__()

        self.config = config
        self.gen = gen
        self.n_active = n_active

        # onehots
        self.active_embedding = nn.Embedding(3, config.entity_embedding_dim)
        self.fainted_embedding = nn.Embedding(3, config.entity_embedding_dim)
        self.gender_embedding = nn.Embedding(
            len(GENDERS) + 1, config.entity_embedding_dim
        )
        self.status_embedding = nn.Embedding(
            len(STATUS) + 1, config.entity_embedding_dim
        )
        self.forme_embedding = nn.Embedding(
            len(TOKENIZED_SCHEMA[f"gen{gen}"]["pokedex"]["forme"]) + 1,
            config.entity_embedding_dim,
        )

        self.hp_sqrt_onehot = nn.Embedding.from_pretrained(sqrt_one_hot_matrix(2048))
        self.hp_embedding = nn.Linear(
            self.hp_sqrt_onehot.embedding_dim + 1,
            config.entity_embedding_dim,
            bias=False,
        )

        self.level_embedding = nn.Embedding(102, config.entity_embedding_dim)

        self.atk_sqrt_onehot = nn.Embedding.from_pretrained(sqrt_one_hot_matrix(1024))
        self.def_sqrt_onehot = nn.Embedding.from_pretrained(sqrt_one_hot_matrix(1024))
        self.spa_sqrt_onehot = nn.Embedding.from_pretrained(sqrt_one_hot_matrix(1024))
        self.spd_sqrt_onehot = nn.Embedding.from_pretrained(sqrt_one_hot_matrix(1024))
        self.spe_sqrt_onehot = nn.Embedding.from_pretrained(sqrt_one_hot_matrix(1024))
        self.stat_embedding = nn.Linear(
            self.atk_sqrt_onehot.embedding_dim
            + self.def_sqrt_onehot.embedding_dim
            + self.spa_sqrt_onehot.embedding_dim
            + self.spd_sqrt_onehot.embedding_dim
            + self.spe_sqrt_onehot.embedding_dim,
            config.entity_embedding_dim,
            bias=False,
        )

        self.pp_sqrt_onehot = nn.Embedding.from_pretrained(sqrt_one_hot_matrix(64))
        self.move_embedding_bin = MoveEmbedding(gen=gen)
        self.move_embedding = nn.Linear(
            self.pp_sqrt_onehot.embedding_dim + self.move_embedding_bin.embedding_dim,
            config.entity_embedding_dim,
            bias=False,
        )

        ability_embedding = AbilityEmbedding(gen=gen)
        self.ability_embedding = nn.Sequential(
            ability_embedding,
            nn.Linear(
                ability_embedding.embedding_dim, config.entity_embedding_dim, bias=False
            ),
        )

        pokedex_embedding = PokedexEmbedding(gen=gen)
        self.pokedex_embedding = nn.Sequential(
            pokedex_embedding,
            nn.Linear(
                pokedex_embedding.embedding_dim, config.entity_embedding_dim, bias=False
            ),
        )

        item_embedding = ItemEmbedding(gen=gen)
        self.item_embedding = nn.Sequential(
            item_embedding,
            nn.Linear(
                item_embedding.embedding_dim, config.entity_embedding_dim, bias=False
            ),
        )

        if gen == 9:
            self.commanding_embedding = nn.Embedding(3, config.entity_embedding_dim)
            self.reviving_embedding = nn.Embedding(3, config.entity_embedding_dim)
            self.tera_embedding = nn.Embedding(3, config.entity_embedding_dim)
            self.teratype_embedding = nn.Embedding(
                len(BattleTypeChart) + 1, config.entity_embedding_dim
            )

        elif gen == 8:
            self.can_gmax_embedding = nn.Embedding(3, config.entity_embedding_dim)

        self.transformer = TransformerEncoder(
            key_size=config.entity_embedding_dim,
            value_size=config.entity_embedding_dim,
            num_heads=config.transformer_num_heads,
            num_layers=config.transformer_num_layers,
            resblocks_num_before=config.resblocks_num_before,
            resblocks_num_after=config.resblocks_num_after,
        )
        self.output = ToVector(
            input_dim=config.entity_embedding_dim,
            output_dim=2 * config.entity_embedding_dim,
        )
        self.moves_w_context = GLU(
            config.entity_embedding_dim,
            config.entity_embedding_dim,
            config.entity_embedding_dim,
        )
        self.spatial = private_spatial.PrivateSpatialEncoder(
            gen=gen, n_active=n_active, config=config
        )

    def embed_hp(self, hp: torch.Tensor, maxhp: torch.Tensor) -> torch.Tensor:
        hp_ratio = (hp / maxhp.clamp(min=1)).unsqueeze(-1)
        hp_sqrt_onehot = self.hp_sqrt_onehot(hp)
        hp_sqrt_onehot = torch.cat([hp_ratio, hp_sqrt_onehot], dim=-1)
        return self.hp_embedding(hp_sqrt_onehot)

    def embed_stat(
        self,
        stat_atk: torch.Tensor,
        stat_def: torch.Tensor,
        stat_spa: torch.Tensor,
        stat_spd: torch.Tensor,
        stat_spe: torch.Tensor,
    ) -> torch.Tensor:
        stat_atk_emb = self.atk_sqrt_onehot(stat_atk)
        stat_def_emb = self.def_sqrt_onehot(stat_def)
        stat_spa_emb = self.spa_sqrt_onehot(stat_spa)
        stat_spd_emb = self.spd_sqrt_onehot(stat_spd)
        stat_spe_emb = self.spe_sqrt_onehot(stat_spe)
        stat_emb = torch.cat(
            [
                stat_atk_emb,
                stat_def_emb,
                stat_spa_emb,
                stat_spd_emb,
                stat_spe_emb,
            ],
            dim=-1,
        )
        return self.stat_embedding(stat_emb)

    def embed_pp(self, pp: torch.Tensor) -> torch.Tensor:
        pp_bin = self.pp_sqrt_onehot(pp)
        return pp_bin

    def embed_moves(
        self, move_tokens: torch.Tensor, move_used: torch.Tensor
    ) -> torch.Tensor:
        moves_emb = self.move_embedding_bin(move_tokens)
        move_used_emb = self.embed_pp(move_used)
        moves_emb = torch.cat((moves_emb, move_used_emb), dim=-1)
        return self.move_embedding(moves_emb)

    def moves_given_context(self, active: torch.Tensor, moves: torch.Tensor):
        active = active.unsqueeze(-2).repeat_interleave(4, -2)
        moves = moves[..., : self.n_active, :, :]
        return self.moves_w_context(moves, active)

    def forward(self, private_reserve: torch.Tensor):
        private_reserve_x = private_reserve + 1
        ability = private_reserve_x[..., 0]
        active = private_reserve_x[..., 1]
        fainted = private_reserve_x[..., 2]
        gender = private_reserve_x[..., 3]
        hp = private_reserve_x[..., 4]
        item = private_reserve_x[..., 5]
        level = private_reserve_x[..., 6]
        maxhp = private_reserve_x[..., 7]
        name = private_reserve_x[..., 8]
        forme = private_reserve_x[..., 9]
        stat_atk = private_reserve_x[..., 10]
        stat_def = private_reserve_x[..., 11]
        stat_spa = private_reserve_x[..., 12]
        stat_spd = private_reserve_x[..., 13]
        stat_spe = private_reserve_x[..., 14]
        status = private_reserve_x[..., 15]

        if self.gen == 9:
            commanding = private_reserve_x[..., 16]
            reviving = private_reserve_x[..., 17]
            teraType = private_reserve_x[..., 18]
            terastallized = private_reserve_x[..., 19]

        elif self.gen == 8:
            canGmax = private_reserve_x[..., 16]

        moves = private_reserve_x[..., -8:]
        moves = moves.view(*moves.shape[:-1], 4, 2)
        move_tokens = moves[..., 0]
        move_used = moves[..., 1]

        ability_emb = self.ability_embedding(ability)
        active_emb = self.active_embedding(active)
        fainted_emb = self.fainted_embedding(fainted)
        gender_emb = self.gender_embedding(gender)
        hp_emb = self.embed_hp(hp, maxhp)
        item_emb = self.item_embedding(item)
        level_emb = self.level_embedding(level)
        name_emb = self.pokedex_embedding(name)
        forme_emb = self.forme_embedding(forme)
        stat_emb = self.embed_stat(stat_atk, stat_def, stat_spa, stat_spd, stat_spe)
        status_emb = self.status_embedding(status)
        moves_emb = self.embed_moves(move_tokens, move_used)

        entity_embeddings = (
            ability_emb
            + active_emb
            + fainted_emb
            + gender_emb
            + hp_emb
            + item_emb
            + level_emb
            + name_emb
            + forme_emb
            + stat_emb
            + status_emb
            + moves_emb.sum(-2)
        )

        if self.gen == 9:
            entity_embeddings += self.commanding_embedding(commanding)
            entity_embeddings += self.reviving_embedding(reviving)
            entity_embeddings += self.teratype_embedding(teraType)
            entity_embeddings += self.tera_embedding(terastallized)

        elif self.gen == 8:
            entity_embeddings += self.can_gmax_embedding(canGmax)

        T, B, S, *_ = entity_embeddings.shape

        mask = (name == 0) | (fainted == 2)
        mask = mask.bool()
        mask = mask.view(T * B, S)

        encoder_mask = mask.clone()
        encoder_mask[..., 0] = False

        entity_embeddings: torch.Tensor = entity_embeddings.view(T * B, S, -1)
        entity_embeddings = self.transformer(entity_embeddings, encoder_mask)

        entity_embeddings = entity_embeddings.view(T * B, S, -1)
        mask = mask.view(T * B, S, -1)

        entity_embeddings = entity_embeddings * ~mask
        entity_embedding = self.output(entity_embeddings)
        entity_embedding = entity_embedding.view(T, B, -1)

        spatial_embedding = self.spatial(entity_embeddings, ~mask)
        spatial_embedding = spatial_embedding.view(T, B, -1)

        entity_embeddings = entity_embeddings.view(T, B, S, -1)

        active = entity_embeddings[..., : self.n_active, :]
        moves = self.moves_given_context(active, moves_emb)

        return entity_embedding, spatial_embedding, moves, entity_embeddings



import math
import torch
import torch.nn as nn

from copy import deepcopy
from typing import Optional, Union, Literal, Tuple, Dict, Any, List

from meloetta.actors.types import State, Choices

from meloetta.frameworks.mewzero.model.utils import _log_policy

from meloetta.frameworks.mewzero.model import config
from meloetta.frameworks.mewzero.model import (
    Encoder,
    PolicyHeads,
    Core,
    ValueHead,
    Indices,
    Logits,
    Policy,
    LogPolicy,
    TrainingOutput,
    EnvStep,
    PostProcess,
    Batch,
)

from meloetta.data import _STATE_FIELDS


class MewZeroModel(nn.Module):
    def __init__(
        self,
        gen: int = 9,
        gametype: Literal["singles", "doubles", "triples"] = "singles",
        config: config.MewZeroModelConfig = config.MewZeroModelConfig(),
    ) -> None:
        super().__init__()

        self.gen = gen
        self.gametype = gametype
        n_active = {"singles": 1, "doubles": 2, "triples": 3}[gametype]

        self.encoder = Encoder(gen=gen, n_active=n_active, config=config.encoder_config)
        self.policy_heads = PolicyHeads(
            gen=gen, gametype=gametype, config=config.policy_heads_config
        )

        self.state_fields = deepcopy(_STATE_FIELDS)

        if gen != 8:
            self.state_fields.remove("max_move_mask")

        if self.gametype == "singles":
            self.state_fields.remove("target_mask")
            self.state_fields.remove("prev_choices")

        self.core = Core(config.resnet_core_config)
        self.value_Head = ValueHead(config.value_head_config)

    def get_learnable_params(self) -> int:
        return sum([p.numel() for p in self.parameters() if p.requires_grad])

    def clean(self, state: State) -> State:
        return {k: state[k] for k in self.state_fields}

    def forward(
        self,
        state: State,
        hidden_state: Tuple[torch.Tensor, torch.Tensor],
        choices: Choices = None,
        training: bool = False,
        need_log_policy: bool = True,
    ):
        encoder_output = self.encoder.forward(state)
        state_emb, hidden_state = self.core.forward(encoder_output, hidden_state)
        value = self.value_Head.forward(state_emb)
        indices, logits, policy = self.policy_heads.forward(
            state_emb,
            encoder_output,
            state,
        )
        return self._postprocess(
            state,
            hidden_state,
            indices,
            logits,
            policy,
            value,
            choices,
            training,
            need_log_policy,
        )

    def learning_forward(
        self,
        batch: Batch,
        hidden_state: Tuple[torch.Tensor, torch.Tensor],
    ):
        T, B = batch.trajectory_length, batch.batch_size

        outputs: List[TrainingOutput] = []

        minibatch_size = min(128, (4096 // batch.trajectory_length) + 1)

        for batch_index in range(math.ceil(batch.batch_size / minibatch_size)):
            mini_state_h, mini_state_c = hidden_state

            batch_start = minibatch_size * batch_index
            batch_end = minibatch_size * (batch_index + 1)

            mini_state_h = mini_state_h[:, batch_start:batch_end].contiguous()
            mini_state_c = mini_state_c[:, batch_start:batch_end].contiguous()

            minibatch = {
                key: value[:, batch_start:batch_end].contiguous()
                for key, value in batch._asdict().items()
                if value is not None
            }
            output, (mini_state_h, mini_state_c) = self.forward(
                minibatch,
                (mini_state_h, mini_state_c),
                training=True,
            )
            outputs.append(output)

        pi_dict: Dict[str, List[torch.Tensor]] = {}
        log_pi_dict: Dict[str, List[torch.Tensor]] = {}
        logit_dict: Dict[str, List[torch.Tensor]] = {}
        vs: List[torch.Tensor] = []

        for o in outputs:
            for key in Policy._fields:
                mbv = getattr(o.pi, key)
                if getattr(o.pi, key) is not None:
                    if key not in pi_dict:
                        pi_dict[key] = []
                    pi_dict[key].append(mbv)

            for key in LogPolicy._fields:
                mbv = getattr(o.log_pi, key)
                if getattr(o.log_pi, key) is not None:
                    if key not in log_pi_dict:
                        log_pi_dict[key] = []
                    log_pi_dict[key].append(mbv)

            for key in Logits._fields:
                mbv = getattr(o.logit, key)
                if getattr(o.logit, key) is not None:
                    if key not in logit_dict:
                        logit_dict[key] = []
                    logit_dict[key].append(mbv)

            vs.append(o.v)

        pi_dict = {
            key: torch.cat(value, dim=1)
            .contiguous()
            .view(T, B, *value[0].shape[2:])
            .squeeze()
            for key, value in pi_dict.items()
        }
        log_pi_dict = {
            key: torch.cat(value, dim=1)
            .contiguous()
            .view(T, B, *value[0].shape[2:])
            .squeeze()
            for key, value in log_pi_dict.items()
        }
        logit_dict = {
            key: torch.cat(value, dim=1)
            .contiguous()
            .view(T, B, *value[0].shape[2:])
            .squeeze()
            for key, value in logit_dict.items()
        }
        vs = torch.cat(vs, dim=1).contiguous().view(T, B, 1)

        return TrainingOutput(
            pi=Policy(**pi_dict),
            v=vs,
            logit=Logits(**logit_dict),
            log_pi=LogPolicy(**log_pi_dict),
        )

    def _postprocess(
        self,
        state: State,
        hidden_state: Tuple[torch.Tensor, torch.Tensor],
        indices: Indices,
        logits: Logits,
        policy: Policy,
        value: torch.Tensor,
        choices: Optional[Dict[str, Any]] = None,
        training: bool = False,
        need_log_policy: bool = True,
    ) -> Union[
        Tuple[TrainingOutput, Tuple[torch.Tensor, torch.Tensor]],
        Tuple[EnvStep, PostProcess, Tuple[torch.Tensor, torch.Tensor]],
    ]:
        if not training:
            targeting = bool(state.get("targeting"))
            post_process = self.index_to_action(targeting, indices, choices)
            env_step = EnvStep(
                indices=indices,
                policy=policy,
                logits=logits,
            )
            output = (env_step, post_process, hidden_state)
        else:
            if need_log_policy:
                log_policy = self.get_log_policy(logits, state)
            else:
                log_policy = None
            output = (
                TrainingOutput(
                    pi=policy,
                    v=value,
                    log_pi=log_policy,
                    logit=logits,
                ),
                hidden_state,
            )
        return output

    def get_log_policy(self, logits: Logits, state: State):
        if self.gametype != "singles":
            target_log_policy = _log_policy(
                logits.target_logits,
                state["target_mask"],
            )
        else:
            target_log_policy = None

        if self.gen == 8:
            max_move_log_policy = _log_policy(
                logits.max_move_logits,
                state["max_move_mask"],
            )
        else:
            max_move_log_policy = None

        if self.gen >= 6:
            flag_log_policy = _log_policy(
                logits.flag_logits,
                state["flag_mask"],
            )
        else:
            flag_log_policy = None

        return LogPolicy(
            action_type_log_policy=_log_policy(
                logits.action_type_logits,
                state["action_type_mask"],
            ),
            move_log_policy=_log_policy(
                logits.move_logits,
                state["move_mask"],
            ),
            switch_log_policy=_log_policy(
                logits.switch_logits,
                state["switch_mask"],
            ),
            max_move_log_policy=max_move_log_policy,
            flag_log_policy=flag_log_policy,
            target_log_policy=target_log_policy,
        )

    def index_to_action(
        self,
        targeting: bool,
        indices: Indices,
        choices: Optional[Dict[str, Any]] = None,
    ):
        action_type_index = indices.action_type_index
        move_index = indices.move_index
        max_move_index = indices.max_move_index
        switch_index = indices.switch_index
        flag_index = indices.flag_index
        target_index = indices.target_index

        if not targeting:
            index = action_type_index

            if action_type_index == 0:
                if flag_index == 3:
                    index = max_move_index
                else:
                    index = move_index

                if choices is not None:
                    if flag_index == 0:
                        data = choices["moves"]
                    elif flag_index == 1:
                        data = choices["mega_moves"]
                    elif flag_index == 2:
                        data = choices["zmoves"]
                    elif flag_index == 3:
                        data = choices["max_moves"]
                    elif flag_index == 4:
                        data = choices["tera_moves"]
                    else:
                        data = choices["moves"]
                else:
                    data = None

            elif action_type_index == 1:
                if choices is not None:
                    data = choices["switches"]
                else:
                    data = None
                index = switch_index

        else:
            if choices is not None:
                data = choices["targets"]
            else:
                data = None
            index = target_index

        return PostProcess(data, index)



import math
import torch

from typing import NamedTuple, Dict, List

from meloetta.actors.types import State, Choices


class SideEncoderOutput(NamedTuple):
    pokemon_embedding: torch.Tensor
    boosts: torch.Tensor
    volatiles: torch.Tensor
    side_conditions: torch.Tensor
    moves: torch.Tensor
    switches: torch.Tensor


class EncoderOutput(NamedTuple):
    pokemon_embedding: torch.Tensor
    boosts: torch.Tensor
    volatiles: torch.Tensor
    side_conditions: torch.Tensor
    moves: torch.Tensor
    switches: torch.Tensor
    weather_emb: torch.Tensor
    scalar_emb: torch.Tensor


class Indices(NamedTuple):
    action_type_index: torch.Tensor = None
    move_index: torch.Tensor = None
    max_move_index: torch.Tensor = None
    switch_index: torch.Tensor = None
    flag_index: torch.Tensor = None
    target_index: torch.Tensor = None

    @classmethod
    def from_list(self, outputs: List[Dict[str, torch.Tensor]]):
        buckets = {k: [] for k in self._fields}
        for output in outputs:
            for key in buckets:
                value = getattr(output, key, None)
                if value is not None:
                    buckets[key].append(value)
        for key, value in buckets.items():
            if buckets[key]:
                buckets[key] = torch.cat(buckets[key], dim=1)
            else:
                buckets[key] = None
        return self(**buckets)

    def to_json(self):
        return {
            k: v.squeeze().tolist() for k, v in self._asdict().items() if v is not None
        }

    def to(self, device):
        return Indices(
            **{k: v.to(device) for k, v in self._asdict().items() if v is not None}
        )

    def flatten(self, *args, **kwargs) -> Dict[str, torch.Tensor]:
        return {
            k: v.flatten(*args, *kwargs).unsqueeze(0)
            for k, v in self._asdict().items()
            if v is not None
        }

    def flatten_without_padding(self, t: torch.Tensor):
        return {
            k: torch.cat([v[: t[i], i] for i in range(v.shape[1])]).unsqueeze(0)
            for k, v in self._asdict().items()
            if v is not None
        }

    def view(self, t, b):
        return Indices(
            **{
                k: v.view(t, b, *v.shape[2:])
                for k, v in self._asdict().items()
                if v is not None
            }
        )


class Policy(NamedTuple):
    action_type: torch.Tensor = None
    move: torch.Tensor = None
    max_move: torch.Tensor = None
    switch: torch.Tensor = None
    flag: torch.Tensor = None
    target: torch.Tensor = None

    @classmethod
    def from_list(self, outputs: List[Dict[str, torch.Tensor]]):
        buckets = {k: [] for k in self._fields}
        for output in outputs:
            for key in buckets:
                value = getattr(output, key, None)
                if value is not None:
                    buckets[key].append(value)
        for key, value in buckets.items():
            if buckets[key]:
                buckets[key] = torch.cat(buckets[key], dim=1)
            else:
                buckets[key] = None
        return self(**buckets)

    def to_json(self):
        return {
            k: v.squeeze().tolist() for k, v in self._asdict().items() if v is not None
        }

    def to(self, device):
        return Policy(
            **{k: v.to(device) for k, v in self._asdict().items() if v is not None}
        )

    def view(self, t, b):
        return Policy(
            **{
                k: v.view(t, b, *v.shape[2:])
                for k, v in self._asdict().items()
                if v is not None
            }
        )


class ModelOutput(NamedTuple):
    pi: Policy
    logit: Policy
    indices: Indices
    v: torch.Tensor = None
    log_pi: Policy = None

    state_emb: torch.Tensor = None
    state_action_emb: torch.Tensor = None

    @classmethod
    def from_list(self, outputs: List[Dict[str, torch.Tensor]]):
        buckets = {k: [] for k in self._fields}
        for output in outputs:
            for key in buckets:
                value = getattr(output, key, None)
                if value is not None:
                    buckets[key].append(value)
        buckets["pi"] = Policy.from_list(buckets["pi"])
        buckets["logit"] = Policy.from_list(buckets["logit"])
        buckets["log_pi"] = Policy.from_list(buckets["log_pi"])
        buckets["indices"] = Indices.from_list(buckets["indices"])
        buckets["v"] = torch.cat(buckets["v"], dim=1).unsqueeze(-1)
        return self(**buckets)

    def to_store(self, state: State) -> Dict[str, torch.Tensor]:
        to_store = {k: v for k, v in state.items() if isinstance(v, torch.Tensor)}
        to_store.update(
            {k: v.squeeze() for k, v in self.indices._asdict().items() if v is not None}
        )
        to_store.update(
            {
                k + "_policy": v.squeeze()
                for k, v in self.pi._asdict().items()
                if v is not None
            }
        )
        return to_store

    def to(self, device):
        return ModelOutput(
            **{k: v.to(device) for k, v in self._asdict().items() if v is not None}
        )

    def view(self, t, b):
        return ModelOutput(
            pi=self.pi.view(t, b),
            logit=self.logit.view(t, b),
            log_pi=self.log_pi.view(t, b),
            indices=self.indices.view(t, b),
            v=self.v.view(t, b, -1),
        )


class PostProcess(NamedTuple):
    data: Choices
    index: torch.Tensor


class Batch(NamedTuple):
    sides: torch.Tensor = None
    boosts: torch.Tensor = None
    volatiles: torch.Tensor = None
    side_conditions: torch.Tensor = None
    pseudoweathers: torch.Tensor = None
    weather: torch.Tensor = None
    wisher: torch.Tensor = None
    scalars: torch.Tensor = None
    turns_since_last_move: torch.Tensor = None
    action_type_mask: torch.Tensor = None
    move_mask: torch.Tensor = None
    max_move_mask: torch.Tensor = None
    switch_mask: torch.Tensor = None
    flag_mask: torch.Tensor = None
    rewards: torch.Tensor = None
    player_id: torch.Tensor = None
    action_type_policy: torch.Tensor = None
    move_policy: torch.Tensor = None
    max_move_policy: torch.Tensor = None
    switch_policy: torch.Tensor = None
    flag_policy: torch.Tensor = None
    action_type_index: torch.Tensor = None
    move_index: torch.Tensor = None
    max_move_index: torch.Tensor = None
    switch_index: torch.Tensor = None
    flag_index: torch.Tensor = None
    value: torch.Tensor = None
    valid: torch.Tensor = None

    @property
    def batch_size(self):
        return self.valid.shape[1]

    @property
    def trajectory_lengths(self):
        return self.valid.sum(0)

    @property
    def trajectory_length(self):
        return self.valid.shape[0]

    def to(self, device: str, non_blocking: bool = False):
        batch = {
            k: t.to(
                device=device,
                non_blocking=non_blocking,
            )
            for k, t in self._asdict().items()
            if t is not None
        }
        return Batch(**batch)

    def flatten(self, *args, **kwargs) -> Dict[str, torch.Tensor]:
        return {
            k: v.flatten(*args, *kwargs).unsqueeze(0)
            for k, v in self._asdict().items()
            if v is not None
        }

    def flatten_without_padding(self, t: torch.Tensor):
        return {
            k: torch.cat([v[: t[i], i] for i in range(v.shape[1])]).unsqueeze(0)
            for k, v in self._asdict().items()
            if v is not None
        }


class Loss(NamedTuple):
    action_type_policy_loss: float = 0
    move_policy_loss: float = 0
    max_move_policy_loss: float = 0
    switch_policy_loss: float = 0
    flag_policy_loss: float = 0
    target_policy_loss: float = 0

    value_loss: float = 0
    repr_loss: float = 0


class Targets(NamedTuple):
    value_targets: List[torch.Tensor] = None
    has_played: List[torch.Tensor] = None

    action_type_policy_target: torch.Tensor = None
    move_policy_target: torch.Tensor = None
    switch_policy_target: torch.Tensor = None
    max_move_policy_target: torch.Tensor = None
    flag_policy_target: torch.Tensor = None
    target_policy_target: torch.Tensor = None

    action_type_is: torch.Tensor = None
    move_is: torch.Tensor = None
    switch_is: torch.Tensor = None
    max_move_is: torch.Tensor = None
    flag_is: torch.Tensor = None
    target_is: torch.Tensor = None

    @property
    def batch_size(self):
        return self.value_targets[0].shape[1]

    @property
    def trajectory_length(self):
        return self.value_targets[0].shape[0]

    def flatten(self, *args, **kwargs) -> Dict[str, List[torch.Tensor]]:
        return {
            k: [
                [v.flatten(*args, *kwargs).unsqueeze(0) for v in vi]
                if isinstance(vi, list)
                else vi.flatten(*args, *kwargs).unsqueeze(0)
                for vi in vl
            ]
            for k, vl in self._asdict().items()
            if vl is not None
        }

    def flatten_without_padding(self, t: torch.Tensor):
        return {
            k: [
                torch.cat([v[: t[i], i] for i in range(v.shape[1])]).unsqueeze(0)
                for v in vt
            ]
            for k, vt in self._asdict().items()
            if vt is not None
        }

    def to(self, device: str, non_blocking: bool = False):
        batch = {
            k: t.to(
                device=device,
                non_blocking=non_blocking,
            )
            for k, t in self._asdict().items()
            if t is not None
        }



import torch

from typing import Dict, List

from meloetta.data import CHOICE_FLAGS


def to_id(string: str):
    return "".join(c for c in string if c.isalnum()).lower()


def _get_private_reserve_size(gen: int):
    if gen == 9:
        return 28
    elif gen == 8:
        return 25
    else:
        return 24


def get_buffer_specs(
    trajectory_length: int,
    gen: int,
    gametype: str,
    private_reserve_size: int,
):
    if gametype != "singles":
        if gametype == "doubles":
            n_active = 2
        else:
            n_active = 3
    else:
        n_active = 1

    buffer_specs = {
        "private_reserve": {
            "size": (trajectory_length, 6, private_reserve_size),
            "dtype": torch.int64,
        },
        "public_n": {
            "size": (trajectory_length, 2),
            "dtype": torch.int64,
        },
        "public_total_pokemon": {
            "size": (trajectory_length, 2),
            "dtype": torch.int64,
        },
        "public_faint_counter": {
            "size": (trajectory_length, 2),
            "dtype": torch.int64,
        },
        "public_side_conditions": {
            "size": (trajectory_length, 2, 18, 3),
            "dtype": torch.int64,
        },
        "public_wisher": {
            "size": (trajectory_length, 2),
            "dtype": torch.int64,
        },
        "public_active": {
            "size": (trajectory_length, 2, n_active, 158),
            "dtype": torch.int64,
        },
        "public_reserve": {
            "size": (trajectory_length, 2, 6, 37),
            "dtype": torch.int64,
        },
        "public_stealthrock": {
            "size": (trajectory_length, 2),
            "dtype": torch.int64,
        },
        "public_spikes": {
            "size": (trajectory_length, 2),
            "dtype": torch.int64,
        },
        "public_toxicspikes": {
            "size": (trajectory_length, 2),
            "dtype": torch.int64,
        },
        "public_stickyweb": {
            "size": (trajectory_length, 2),
            "dtype": torch.int64,
        },
        "weather": {
            "size": (trajectory_length,),
            "dtype": torch.int64,
        },
        "weather_time_left": {
            "size": (trajectory_length,),
            "dtype": torch.int64,
        },
        "weather_min_time_left": {
            "size": (trajectory_length,),
            "dtype": torch.int64,
        },
        "pseudo_weather": {
            "size": (trajectory_length, 12, 2),
            "dtype": torch.int64,
        },
        "turn": {
            "size": (trajectory_length,),
            "dtype": torch.int64,
        },
        "turns_since_last_move": {
            "size": (trajectory_length,),
            "dtype": torch.int64,
        },
        "action_type_mask": {
            "size": (trajectory_length, 3),
            "dtype": torch.bool,
        },
        "move_mask": {
            "size": (trajectory_length, 4),
            "dtype": torch.bool,
        },
        "switch_mask": {
            "size": (trajectory_length, 6),
            "dtype": torch.bool,
        },
        "flag_mask": {
            "size": (trajectory_length, len(CHOICE_FLAGS)),
            "dtype": torch.bool,
        },
        "rewards": {
            "size": (trajectory_length, 2),
            "dtype": torch.float32,
        },
        "player_id": {
            "size": (trajectory_length,),
            "dtype": torch.long,
        },
    }
    if gametype != "singles":
        buffer_specs.update(
            {
                "target_mask": {
                    "size": (trajectory_length, 4),
                    "dtype": torch.bool,
                },
                "prev_choices": {
                    "size": (trajectory_length, 2, 4),
                    "dtype": torch.int64,
                },
                "choices_done": {
                    "size": (trajectory_length, 1),
                    "dtype": torch.int64,
                },
                "targeting": {
                    "size": (trajectory_length, 1),
                    "dtype": torch.int64,
                },
                "target_policy": {
                    "size": (trajectory_length, 2 * n_active),
                    "dtype": torch.float32,
                },
                "target_index": {
                    "size": (trajectory_length,),
                    "dtype": torch.int64,
                },
            }
        )

    # add policy...
    buffer_specs.update(
        {
            "action_type_policy": {
                "size": (trajectory_length, 3),
                "dtype": torch.float32,
            },
            "move_policy": {
                "size": (trajectory_length, 4),
                "dtype": torch.float32,
            },
            "switch_policy": {
                "size": (trajectory_length, 6),
                "dtype": torch.float32,
            },
            "flag_policy": {
                "size": (trajectory_length, len(CHOICE_FLAGS)),
                "dtype": torch.float32,
            },
        }
    )

    # ...and indices
    buffer_specs.update(
        {
            "action_type_index": {
                "size": (trajectory_length,),
                "dtype": torch.int64,
            },
            "move_index": {
                "size": (trajectory_length,),
                "dtype": torch.int64,
            },
            "switch_index": {
                "size": (trajectory_length,),
                "dtype": torch.int64,
            },
            "flag_index": {
                "size": (trajectory_length,),
                "dtype": torch.int64,
            },
        }
    )

    if gen == 8:
        buffer_specs.update(
            {
                "max_move_mask": {
                    "size": (trajectory_length, 4),
                    "dtype": torch.bool,
                },
                "max_move_policy": {
                    "size": (trajectory_length, 4),
                    "dtype": torch.float32,
                },
                "max_move_index": {
                    "size": (trajectory_length,),
                    "dtype": torch.int64,
                },
            }
        )

    buffer_specs.update(
        {
            "valid": {
                "size": (trajectory_length,),
                "dtype": torch.bool,
            }
        }
    )
    return buffer_specs


def create_buffers(num_buffers: int, trajectory_length: int, gen: int, gametype: str):
    """
    num_buffers: int
        the size of the replay buffer
    trajectory_length: int
        the max length of a replay
    battle_format: str
        the type of format
    """

    private_reserve_size = _get_private_reserve_size(gen)

    buffer_specs = get_buffer_specs(
        trajectory_length, gen, gametype, private_reserve_size
    )

    buffers: Dict[str, List[torch.Tensor]] = {key: [] for key in buffer_specs}
    for _ in range(num_buffers):
        for key in buffer_specs:
            if key.endswith("_mask"):
                buffers[key].append(torch.ones(**buffer_specs[key]).share_memory_())
            else:
                buffers[key].append(torch.zeros(**buffer_specs[key]).share_memory_())
    return buffers


def expand_bt(tensor: torch.Tensor, time: int = 1, batch: int = 1) -> torch.Tensor:
    shape = tensor.shape
    return tensor.view(time, batch, *(tensor.shape or shape))



import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

import random

from tqdm import tqdm


from meloetta.frameworks.nash_ketchum.model.config import EncoderConfig
from meloetta.frameworks.nash_ketchum.model.encoders import (
    PrivateEncoder,
    PublicEncoder,
    ScalarEncoder,
    WeatherEncoder,
)

from meloetta.data import (
    GENDERS,
)


class Siamese(nn.Module):
    def __init__(self, gen: int, n_active: int, config: EncoderConfig) -> None:
        super().__init__()

        self.private_encoder = PrivateEncoder(
            gen=gen, n_active=n_active, config=config.private_encoder_config
        )
        self.public_encoder = PublicEncoder(
            gen=gen, n_active=n_active, config=config.public_encoder_config
        )
        self.weather_encoder = WeatherEncoder(
            gen=gen, config=config.weather_encoder_config
        )
        self.scalar_encoder = ScalarEncoder(
            gen=gen, n_active=n_active, config=config.scalar_encoder_config
        )
        moves_dim = EncoderConfig.private_encoder_config.entity_embedding_dim
        self.moves = nn.Sequential(
            nn.Linear(moves_dim, moves_dim),
            nn.ReLU(),
            nn.Linear(moves_dim, moves_dim),
        )
        switches_dim = EncoderConfig.private_encoder_config.entity_embedding_dim
        self.switches = nn.Sequential(
            nn.Linear(switches_dim, switches_dim),
            nn.ReLU(),
            nn.Linear(switches_dim, switches_dim),
        )
        private_entity_emb_dim = (
            EncoderConfig.private_encoder_config.entity_embedding_dim
        )
        self.private_entity_emb = nn.Sequential(
            nn.Linear(private_entity_emb_dim, private_entity_emb_dim),
            nn.ReLU(),
            nn.Linear(private_entity_emb_dim, private_entity_emb_dim),
        )
        public_entity_emb_dim = EncoderConfig.public_encoder_config.entity_embedding_dim
        self.public_entity_emb = nn.Sequential(
            nn.Linear(public_entity_emb_dim, public_entity_emb_dim),
            nn.ReLU(),
            nn.Linear(public_entity_emb_dim, public_entity_emb_dim),
        )
        public_scalar_emb_dim = EncoderConfig.scalar_encoder_config.embedding_dim
        self.public_scalar_emb = nn.Sequential(
            nn.Linear(public_scalar_emb_dim, public_scalar_emb_dim),
            nn.ReLU(),
            nn.Linear(public_scalar_emb_dim, public_scalar_emb_dim),
        )
        weather_emb_dim = EncoderConfig.weather_encoder_config.embedding_dim
        self.weather_emb = nn.Sequential(
            nn.Linear(weather_emb_dim, weather_emb_dim),
            nn.ReLU(),
            nn.Linear(weather_emb_dim, weather_emb_dim),
        )
        scalar_emb_dim = EncoderConfig.scalar_encoder_config.embedding_dim
        self.scalar_emb = nn.Sequential(
            nn.Linear(scalar_emb_dim, scalar_emb_dim),
            nn.ReLU(),
            nn.Linear(scalar_emb_dim, scalar_emb_dim),
        )

    def sample_(self, batch_size: int) -> torch.Tensor:

        team_shape = (batch_size, 7)
        batch = torch.zeros(1, *team_shape, 28, dtype=torch.long)

        batch[..., 0] = torch.randint(
            0,
            self.private_encoder.ability_embedding._modules["0"].num_embeddings - 1,
            team_shape,
        )
        batch[..., 1] = torch.randint(0, 2, team_shape)
        batch[..., 2] = torch.randint(0, 2, team_shape)
        batch[..., 3] = torch.randint(0, len(GENDERS), team_shape)
        batch[..., 4] = torch.randint(0, 2047, team_shape)
        batch[..., 5] = torch.randint(
            0,
            self.private_encoder.item_embedding._modules["0"].num_embeddings - 1,
            team_shape,
        )
        batch[..., 6] = torch.randint(
            0, self.private_encoder.level_sqrt_onehot.num_embeddings - 1, team_shape
        )
        batch[..., 7] = torch.randint(0, 2047, team_shape)
        batch[..., 4] = batch[..., 4].clamp(max=batch[..., 7])
        batch[..., 8] = torch.randint(
            0,
            self.private_encoder.pokedex_embedding._modules["0"].num_embeddings - 1,
            team_shape,
        )
        batch[..., 9] = torch.randint(
            0, self.private_encoder.forme_embedding.num_embeddings - 1, team_shape
        )
        batch[..., 10] = torch.randint(
            0, self.private_encoder.atk_sqrt_onehot.num_embeddings - 1, team_shape
        )
        batch[..., 11] = torch.randint(
            0, self.private_encoder.def_sqrt_onehot.num_embeddings - 1, team_shape
        )
        batch[..., 12] = torch.randint(
            0, self.private_encoder.spa_sqrt_onehot.num_embeddings - 1, team_shape
        )
        batch[..., 13] = torch.randint(
            0, self.private_encoder.spd_sqrt_onehot.num_embeddings - 1, team_shape
        )
        batch[..., 14] = torch.randint(
            0, self.private_encoder.spe_sqrt_onehot.num_embeddings - 1, team_shape
        )
        batch[..., 15] = torch.randint(
            0, self.private_encoder.status_onehot.num_embeddings - 1, team_shape
        )
        batch[..., 16] = torch.randint(
            0, self.private_encoder.commanding_onehot.num_embeddings - 1, team_shape
        )
        batch[..., 17] = torch.randint(
            0, self.private_encoder.reviving_onehot.num_embeddings - 1, team_shape
        )
        batch[..., 18] = torch.randint(
            0, self.private_encoder.teratype_onehot.num_embeddings - 1, team_shape
        )
        batch[..., 19] = torch.randint(
            0, self.private_encoder.tera_onehot.num_embeddings - 1, team_shape
        )
        batch[..., -8::2] = torch.randint(
            0, self.private_encoder.move_embedding.num_embeddings - 1, team_shape + (4,)
        )
        batch[..., -7::2] = torch.randint(0, 63, team_shape + (4,))
        return batch.detach()

    def transform_(self, batch: torch.Tensor, chance: float) -> torch.Tensor:
        batch_size = batch.shape[1]

        team_shape = (1,)
        for bi in range(batch_size):
            for ti in range(batch.shape[2]):
                for i in range(21):
                    c = random.random()
                    if c <= chance:
                        if i == 0:
                            batch[..., bi, ti, 0] = torch.randint(
                                0,
                                self.private_encoder.ability_embedding._modules[
                                    "0"
                                ].num_embeddings
                                - 1,
                                team_shape,
                            )
                        if i == 1:
                            batch[..., bi, ti, 1] = torch.randint(0, 2, team_shape)
                        if i == 2:
                            batch[..., bi, ti, 2] = torch.randint(0, 2, team_shape)
                        if i == 2:
                            batch[..., bi, ti, 3] = torch.randint(
                                0, len(GENDERS), team_shape
                            )
                        if i == 3:
                            batch[..., bi, ti, 4] = torch.randint(0, 2047, team_shape)
                        if i == 4:
                            batch[..., bi, ti, 5] = torch.randint(
                                0,
                                self.private_encoder.item_embedding._modules[
                                    "0"
                                ].num_embeddings
                                - 1,
                                team_shape,
                            )
                        if i == 5:
                            batch[..., bi, ti, 6] = torch.randint(
                                0,
                                self.private_encoder.level_sqrt_onehot.num_embeddings
                                - 1,
                                team_shape,
                            )
                        if i == 6:
                            batch[..., bi, ti, 7] = torch.randint(0, 2047, team_shape)
                            batch[..., bi, ti, 4] = batch[..., bi, ti, 4].clamp(
                                max=batch[..., bi, ti, 7]
                            )
                        if i == 7:
                            batch[..., bi, ti, 8] = torch.randint(
                                0,
                                self.private_encoder.pokedex_embedding._modules[
                                    "0"
                                ].num_embeddings
                                - 1,
                                team_shape,
                            )
                        if i == 8:
                            batch[..., bi, ti, 9] = torch.randint(
                                0,
                                self.private_encoder.forme_embedding.num_embeddings - 1,
                                team_shape,
                            )
                        if i == 10:
                            batch[..., bi, ti, 10] = torch.randint(
                                0,
                                self.private_encoder.atk_sqrt_onehot.num_embeddings - 1,
                                team_shape,
                            )
                        if i == 11:
                            batch[..., bi, ti, 11] = torch.randint(
                                0,
                                self.private_encoder.def_sqrt_onehot.num_embeddings - 1,
                                team_shape,
                            )
                        if i == 12:
                            batch[..., bi, ti, 12] = torch.randint(
                                0,
                                self.private_encoder.spa_sqrt_onehot.num_embeddings - 1,
                                team_shape,
                            )
                        if i == 13:
                            batch[..., bi, ti, 13] = torch.randint(
                                0,
                                self.private_encoder.spd_sqrt_onehot.num_embeddings - 1,
                                team_shape,
                            )
                        if i == 14:
                            batch[..., bi, ti, 14] = torch.randint(
                                0,
                                self.private_encoder.spe_sqrt_onehot.num_embeddings - 1,
                                team_shape,
                            )
                        if i == 15:
                            batch[..., bi, ti, 15] = torch.randint(
                                0,
                                self.private_encoder.status_onehot.num_embeddings - 1,
                                team_shape,
                            )
                        if i == 16:
                            batch[..., bi, ti, 16] = torch.randint(
                                0,
                                self.private_encoder.commanding_onehot.num_embeddings
                                - 1,
                                team_shape,
                            )
                        if i == 17:
                            batch[..., bi, ti, 17] = torch.randint(
                                0,
                                self.private_encoder.reviving_onehot.num_embeddings - 1,
                                team_shape,
                            )
                        if i == 18:
                            batch[..., bi, ti, 18] = torch.randint(
                                0,
                                self.private_encoder.teratype_onehot.num_embeddings - 1,
                                team_shape,
                            )
                        if i == 19:
                            batch[..., bi, ti, 19] = torch.randint(
                                0,
                                self.private_encoder.tera_onehot.num_embeddings - 1,
                                team_shape,
                            )
                        if i == 20:
                            batch[..., bi, ti, -8::2] = torch.randint(
                                0,
                                self.private_encoder.move_embedding.num_embeddings - 1,
                                team_shape + (4,),
                            )
                        if i == 21:
                            batch[..., bi, ti, -7::2] = torch.randint(
                                0, 63, team_shape + (4,)
                            )
        return batch.detach()

    def forward(self, x1: torch.Tensor, x2: torch.Tensor):
        """
        Input:
            x1: first views of images
            x2: second views of images
        Output:
            p1, p2, z1, z2: predictors and targets of the network
            See Sec. 3 of https://arxiv.org/abs/2011.10566 for detailed notations
        """

        # compute features for one view
        z11, z12, z13 = self.encoder(x1)  # NxC
        z21, z22, z23 = self.encoder(x2)  # NxC

        p11 = self.predictor1(z11)  # NxC
        p12 = self.predictor2(z12)  # NxC
        p13 = self.predictor3(z13)  # NxC

        p21 = self.predictor1(z21)  # NxC
        p22 = self.predictor2(z22)  # NxC
        p23 = self.predictor3(z23)  # NxC

        return (
            p11,
            p12,
            p13,
            z11,
            z12,
            z13,
            p21,
            p22,
            p23,
            z21,
            z22,
            z23,
        )


def loss_fn(p: torch.Tensor, z: torch.Tensor) -> torch.Tensor:
    z = z.detach()

    p = F.normalize(p, dim=1)
    z = F.normalize(z, dim=1)
    return -(p * z).sum(dim=1).mean()


def main():
    batch_size = 16
    chance = 0.5
    device = "cuda"

    model = Siamese()
    model.to(device)

    optimizer = optim.Adam(model.parameters(), lr=1e-4)
    criterion = loss_fn

    progress = tqdm(range(1000))
    for _ in progress:
        with torch.no_grad():
            x = model.sample_(batch_size)
            x1 = model.transform_(x.clone(), chance)
            x2 = model.transform_(x.clone(), chance)

        x1 = x1.to(device)
        x2 = x2.to(device)

        (p11, p12, p13, z11, z12, z13, p21, p22, p23, z21, z22, z23) = model(x1, x2)
        p11 = torch.flatten(p11, 0, -2)
        p12 = torch.flatten(p12, 0, -2)
        p13 = torch.flatten(p13, 0, -2)
        z11 = torch.flatten(z11, 0, -2)
        z12 = torch.flatten(z12, 0, -2)
        z13 = torch.flatten(z13, 0, -2)
        p21 = torch.flatten(p21, 0, -2)
        p22 = torch.flatten(p22, 0, -2)
        p23 = torch.flatten(p23, 0, -2)
        z21 = torch.flatten(z21, 0, -2)
        z22 = torch.flatten(z22, 0, -2)
        z23 = torch.flatten(z23, 0, -2)

        loss1 = criterion(p11, z21) + criterion(p21, z11)
        loss2 = criterion(p12, z22) + criterion(p22, z12)
        loss3 = criterion(p13, z23) + criterion(p23, z13)

        loss = (loss1 + loss2 + loss3) / 6

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        progress.set_description(
            f"l1: {loss1.item():.3f}, l2: {loss2.item():.3f}, l3: {loss3.item():.3f}"
        )

    torch.save(
        model.encoder.state_dict(),
        "meloetta/frameworks/nash_ketchum/model/encoders/private_encoder.pt",
    )


if __name__ == "__main__":
    main()



import math
import time
import wandb
import traceback
import threading

import torch
import torch.optim as optim
import torch.nn.functional as F

from meloetta.frameworks.porygon import utils
from meloetta.frameworks.porygon.buffer import ReplayBuffer
from meloetta.frameworks.porygon.config import PorygonConfig
from meloetta.frameworks.porygon.model import (
    PorygonModel,
    TrainingOutput,
    Batch,
    Targets,
    Loss,
    Policy,
)
from meloetta.frameworks.porygon.utils import from_logits as vtrace_from_logits


def compute_baseline_loss(advantages, valid):
    return torch.sum((advantages * valid) ** 2)


def compute_entropy_loss(policy, log_policy, valid):
    """Return the entropy loss, i.e., the negative entropy of the policy."""
    entropy = (policy * log_policy).sum(-1)
    entropy = entropy / (policy > 0).sum(-1).clamp(min=1)
    return torch.sum(entropy * valid)


def compute_policy_gradient_loss(
    logits: torch.Tensor,
    actions: torch.Tensor,
    advantages: torch.Tensor,
    policy_mask: torch.Tensor,
    valid: torch.Tensor,
) -> torch.Tensor:
    logits = torch.flatten(logits, 0, 1)
    policy_mask = torch.flatten(policy_mask, 0, 1)
    masked_logits = torch.masked_fill(logits, ~policy_mask, float("-inf"))
    actions = torch.flatten(actions, 0, 1)
    cross_entropy = F.cross_entropy(
        masked_logits,
        target=actions,
        reduction="none",
    )
    cross_entropy = cross_entropy.view_as(advantages)
    return torch.sum(cross_entropy * advantages.detach() * valid)


class PorygonLearner:
    def __init__(
        self,
        learner_model: PorygonModel,
        actor_model: PorygonModel,
        replay_buffer: ReplayBuffer,
        config: PorygonConfig = PorygonConfig(),
    ):
        self.learner_model = learner_model
        self.actor_model = actor_model

        self.replay_buffer = replay_buffer

        self.config = config

        self.optimizer = optim.Adam(
            self.learner_model.parameters(),
            lr=config.learning_rate,
            betas=(config.adam.b1, config.adam.b2),
            eps=config.adam.eps,
        )
        self.learner_steps = 0

    def get_config(self):
        return {
            "parameters": self.learner_model.get_learnable_params(),
            "learner_steps": self.learner_steps,
            **self.config.__dict__,
        }

    def save(self):
        torch.save(
            {
                "learner_model": self.learner_model.state_dict(),
                "actor_model": self.actor_model.state_dict(),
                "config": self.config,
                "optimizer": self.optimizer.state_dict(),
                "learner_steps": self.learner_steps,
            },
            f"cpkts/cpkt-{self.learner_steps:05}.tar",
        )

    @classmethod
    def from_config(self, config: PorygonConfig = None):
        return PorygonLearner.from_pretrained(config_override=config)

    @classmethod
    def from_pretrained(self, fpath: str = None, config_override: PorygonConfig = None):
        assert not (
            fpath is None and config_override is None
        ), "please set one of `fpath` or `config`"

        if fpath is not None:
            datum = torch.load(fpath)
        else:
            datum = {}

        if datum.get("config"):
            print("Found config!")
            config: PorygonConfig = datum["config"]

        if config_override is not None:
            config = config_override

        gen, gametype = utils.get_gen_and_gametype(config.battle_format)
        learner_model = PorygonModel(
            gen=gen, gametype=gametype, config=config.model_config
        )
        # opt_learner_model = torch.compile(learner_model)
        learner_model.train()

        print(f"learnabled params: {learner_model.get_learnable_params():,}")

        actor_model = PorygonModel(
            gen=gen, gametype=gametype, config=config.model_config
        )
        actor_model.load_state_dict(learner_model.state_dict())
        actor_model.share_memory()
        actor_model.eval()

        if datum.get("learner_model"):
            try:
                learner_model.load_state_dict(datum["learner_model"], strict=False)
                actor_model.load_state_dict(datum["actor_model"], strict=False)
            except Exception as e:
                traceback.print_exc()

        learner_model = learner_model.to(config.learner_device, non_blocking=True)
        actor_model = actor_model.to(config.actor_device, non_blocking=True)

        learner = PorygonLearner(
            learner_model,
            actor_model,
            replay_buffer=ReplayBuffer(
                config.trajectory_length,
                gen,
                gametype,
                config.num_buffers,
                config.learner_device,
            ),
            config=config,
        )

        if datum.get("optimizer"):
            try:
                learner.optimizer.load_state_dict(datum["optimizer"])
            except:
                print("Optimizer not loaded")
        else:
            print("Optimizer not loaded")

        if datum.get("learner_steps"):
            learner.learner_steps = datum["learner_steps"]
        else:
            print("Learner steps not loaded")

        return learner

    def collect_batch_trajectory(self):
        # hidden_state = hidden_state.to(self.config.learner_device, non_blocking=True)
        batch = None
        while batch is None:
            time.sleep(1)
            batch = self.replay_buffer.get_batch(self.config.batch_size)
        return batch

    def step(self):
        batch = self.collect_batch_trajectory()
        self.update_paramters(batch)
        self.learner_steps += 1
        if self.learner_steps % 500 == 0:
            self.save()

    def update_paramters(
        self,
        batch: Batch,
        lock=threading.Lock(),
    ):
        with lock:
            targets = self._get_targets(batch)

            self.optimizer.zero_grad(set_to_none=True)
            losses = self._backpropagate(batch, targets)

            static_loss_dict = losses.to_log(batch)
            static_loss_dict["s"] = self.learner_steps
            wandb.log(static_loss_dict)

            # torch.nn.utils.clip_grad_value_(
            #     self.learner_model.parameters(), self.config.clip_gradient
            # )

            # torch.nn.utils.clip_grad_norm_(
            #     self.learner_model.parameters(), self.config.clip_gradient_norm
            # )

            self.optimizer.step()

            # if self.learner_steps > 10:
            #     if self.learner_steps % 100 == 0:
            #         self.actor_model.load_state_dict(self.learner_model.state_dict())
            # else:

            if self.learner_steps % 20 == 0:
                self.actor_model.load_state_dict(self.learner_model.state_dict())

    def _get_targets(
        self,
        batch: Batch,
    ) -> Targets:
        bidx = torch.arange(
            batch.batch_size, device=next(self.learner_model.parameters()).device
        )
        lengths = batch.valid.sum(0)

        learner: TrainingOutput

        with torch.no_grad():
            learner = self.learner_model.learning_forward(batch)

        bootstrap_value = learner.v.squeeze(-1)[lengths.squeeze() - 1, bidx]

        dones = torch.cat(
            (batch.valid[1:], torch.zeros_like(batch.valid[0, None])), dim=0
        )
        discounts = dones * self.config.gamma

        targets_dict = {}

        for pi_field, pi in learner.pi._asdict().items():
            pi_field: str
            pi: torch.Tensor

            if pi is None:
                continue

            if pi_field == "action_policy":
                policy_mask = torch.cat((batch.move_mask, batch.switch_mask), dim=-1)
            else:
                policy_mask = batch.get_mask_from_policy(pi_field)

            vtrace_returns = vtrace_from_logits(
                behavior_policy_logits=getattr(batch, pi_field),
                target_policy_logits=pi,
                policy_mask=policy_mask,
                actions=batch.get_index_from_policy(pi_field),
                discounts=discounts,
                rewards=batch.rewards,
                values=learner.v.squeeze(-1),
                bootstrap_value=bootstrap_value,
            )

            targets_dict[
                pi_field.replace("_policy", "") + "_vtrace_returns"
            ] = vtrace_returns

        return Targets(**targets_dict)
        # return Targets(
        #     **{
        #         k: [vt.flatten(0, 1).unsqueeze(0) for vt in v]
        #         for k, v in targets_dict.items()
        #     }
        # )

    def _backpropagate(
        self,
        batch: Batch,
        targets: Targets,
    ) -> Loss:

        loss_dict = {
            key + "_loss": 0
            for key, value in batch._asdict().items()
            if "_policy" in key and value is not None
        }
        fields = set([field.replace("_policy_loss", "") for field in loss_dict])

        loss_dict.update({key.replace("policy", "value"): 0 for key in loss_dict})
        loss_dict.update({key.replace("policy", "entropy"): 0 for key in loss_dict})

        total_valid = (batch.valid).sum().clamp(min=1).item()
        total_move_valid = (
            (batch.valid * (batch.action_type_index == 0)).sum().clamp(min=1).item()
        )
        total_switch_valid = (
            (batch.valid * (batch.action_type_index == 1)).sum().clamp(min=1).item()
        )

        minibatch_size = 8

        for batch_index in range(math.ceil(batch.batch_size / minibatch_size)):
            batch_start = minibatch_size * batch_index
            batch_end = minibatch_size * (batch_index + 1)

            loss = 0

            minibatch = batch.slice(batch_start, batch_end)

            if minibatch.valid.sum() == 0:
                continue

            targ = targets.slice(batch_start, batch_end)

            output, _, _ = self.learner_model.forward(
                minibatch._asdict(),
                need_log_policy=True,
            )

            for field in fields:
                pi_field = field + "_policy"
                log_pi_field = field + "_log_policy"
                index_field = field + "_index"
                logit_field = field + "_logits"
                vtrace_field = field + "_vtrace_returns"

                pi = getattr(output.pi, pi_field)
                # logit = getattr(output.logit, logit_field)

                if pi is None:
                    continue

                valid = minibatch.valid
                policy_valid = total_valid

                policy_mask = minibatch.get_mask_from_policy(pi_field)

                if field == "move" or field == "flag":
                    valid = valid * (minibatch.action_type_index == 0)
                    policy_valid = total_move_valid

                elif field == "switch":
                    valid = valid * (minibatch.action_type_index == 1)
                    policy_valid = total_switch_valid

                vtrace_returns = getattr(targ, vtrace_field)

                value_loss = 0.25 * compute_baseline_loss(
                    vtrace_returns.vs - output.v.squeeze(-1), valid
                )

                policy_loss = compute_policy_gradient_loss(
                    getattr(output.logit, logit_field),
                    getattr(minibatch, index_field),
                    vtrace_returns.pg_advantages,
                    policy_mask,
                    valid,
                )

                entropy_loss = 5e-1 * compute_entropy_loss(
                    getattr(output.pi, pi_field),
                    getattr(output.log_pi, log_pi_field),
                    valid,
                )

                policy_loss_field = field + "_policy_loss"
                value_loss_field = field + "_value_loss"
                entropy_loss_field = field + "_entropy_loss"

                loss_dict[policy_loss_field] += policy_loss.item()
                loss_dict[value_loss_field] += value_loss.item()
                loss_dict[entropy_loss_field] += entropy_loss.item()

                loss += (value_loss + policy_loss + entropy_loss) / policy_valid

            loss.backward()

        for key in loss_dict:
            if "action_type" in key:
                loss_dict[key] /= total_valid
            elif "move" in key or "flag" in key:
                loss_dict[key] /= total_move_valid
            elif "switch" in key:
                loss_dict[key] /= total_switch_valid

        return Loss(**loss_dict)

    def run(self):
        while True:
            self.step()



import torch
import torch.nn as nn
import torch.nn.functional as F

from meloetta.frameworks.mewzero.model import config
from meloetta.frameworks.mewzero.model.utils import Resblock, GLU
from meloetta.frameworks.mewzero.model.utils import _legal_policy


class ActionTypeHead(nn.Module):
    def __init__(self, config: config.ActionTypeHeadConfig):
        super().__init__()

        self.num_action_types = config.num_action_types
        self.project = nn.Linear(config.state_embedding_dim, config.residual_dim)
        self.resblocks = nn.Sequential(
            *[Resblock(config.residual_dim, use_layer_norm=True) for _ in range(2)]
        )

        self.action_fc = GLU(
            config.residual_dim, config.num_action_types, config.context_dim
        )
        self.fc1 = nn.Linear(config.num_action_types, config.action_map_dim)
        self.fc2 = nn.Linear(config.action_map_dim, config.action_map_dim)

        self.glu1 = GLU(
            config.action_map_dim,
            config.autoregressive_embedding_dim,
            config.context_dim,
        )
        self.glu2 = GLU(
            config.state_embedding_dim,
            config.autoregressive_embedding_dim,
            config.context_dim,
        )
        self.one_hot = nn.Embedding.from_pretrained(torch.eye(3))

    def forward(
        self,
        state: torch.Tensor,
        scalar_context: torch.Tensor,
        action_type_mask: torch.Tensor,
    ):
        T, B, *_ = state.shape

        embedded_state = self.project(state)
        embedded_state = F.relu(self.resblocks(embedded_state))
        action_type_logits = self.action_fc(embedded_state, scalar_context)
        action_type_policy = _legal_policy(action_type_logits, action_type_mask)
        action_type_index = torch.multinomial(action_type_policy.view(T * B, -1), 1)
        action_type_index = action_type_index.view(T, B)

        action_one_hot = self.one_hot(action_type_index.long())
        embedding1 = F.relu(self.fc1(action_one_hot))
        embedding1 = self.fc2(embedding1)
        embedding1 = self.glu1(embedding1, scalar_context)
        embedding2 = self.glu2(state, scalar_context)
        autoregressive_embedding = embedding1 + embedding2

        return (
            action_type_logits,
            action_type_policy,
            action_type_index,
            autoregressive_embedding,
        )
